//Copyright INRIA/CNRS
//
//Neveu, Emilie; Popov, Petr; Hoffmann, Alexandre; Grudinin, Sergei. 2017
//sergei.grudinin@inria.fr
//
//This software is a computer program whose purpose is to compute
//the root mean sum of squared deviations (RMSDs) corresponding 
//to a rigid-body motion in constatnt time. It can also handle flexible  
//motions represented with collective motions vectors such as NMA or PCA.
//
//This software is governed by the CeCILL-B  license (fully BSD-compatible)
//under French law and abiding by the rules of distribution of free software.
//You can  use, modify and/ or redistribute the software under the terms
//of the CeCILL-B license as circulated by CEA, CNRS and INRIA at the
//following URL "http://www.cecill.info".
//
//As a counterpart to the access to the source code and  rights to copy,
//modify and redistribute granted by the license, users are provided only
//with a limited warranty  and the software's author,  the holder of the
//economic rights,  and the successive licensors  have only  limited
//liability. 
//
//In this respect, the user's attention is drawn to the risks associated
//with loading,  using,  modifying and/or developing or reproducing the
//software by the user in light of its specific status of free software,
//that may mean  that it is complicated to manipulate,  and  that  also
//therefore means  that it is reserved for developers  and  experienced
//professionals having in-depth computer knowledge. Users are therefore
//encouraged to load and test the software's suitability as regards their
//requirements in conditions enabling the security of their systems and/or 
//data to be ensured and,  more generally, to use and operate it in the 
//same conditions as regards security. 
//
//The fact that you are presently reading this means that you have had
//knowledge of the CeCILL-B license and that you accept its terms.

#pragma once

#include <cstddef> // needed for the definition of NULL
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <array>
#include <vector>

template <class Vec3>
class RMSD {

public:
    typedef std::array<double,3> vector3;
//    typedef std::array<vector3, 3> mat3;

//	typedef double vector3[3];
//   typedef double mat3[3][3];
    
	private:
		
		//methods
		void computeCOM(const Vec3 *, int N, const double * = NULL);
		void computeInertia(const Vec3 *, int N, const double * = NULL);
		void computeInertiaTwoMolecules(const Vec3 *xyz, const Vec3 *xyz1, int N, const double * w = NULL);
		void computeCrossCoefficients(const Vec3 *xyz, const Vec3 *xyz1, int N, const double *w = NULL);
		void multiplyMatriceTMatrice(const double *m1, const double *m2, double *m);
        void computeCOMModes( int N,  Vec3 **fxyz, const double *w= NULL);
        void computeCrossCOMModes(const Vec3 *xyz, int N,  Vec3 **fxyz , const double *w =NULL);
        void computeCrossModesModes(int N,  Vec3 **fxyz , const double *w =NULL);
    
		//members
		vector3 COM;
		double	Xxx, Xyy, Xzz;

		double	Ixx, Iyy, Izz;
		double	Ixy, Iyz, Izx;

		double  xy, yz, zx;
		double  xm, ym, zm;
		double  xm2, ym2, zm2;
    
        int nM;
        std::vector<vector3> B;
        std::vector<vector3> Dx, Dy, Dz;
        std::vector< std::vector<vector3> > Fcol1, Fcol2, Fcol3;
	
	public:
		RMSD();
		~RMSD();

		//methods
		void init(const Vec3 *xyz, int N, const double *w = NULL);
        void initModes(const Vec3 *xyz, int N,  Vec3 **fxyz, int nModes, const double *w = NULL);
		void init(const Vec3 *xyz, const Vec3 *xyz1, int N, const double *w = NULL);

		double getMSDMatrix (const double *M, const double *T);
        double getMSDMatrix (const double *M, const double *T, const double *modes1, const double *modes2 = NULL);
        double getMSDOrthoMatrix (const double *M, const double *T, const double *modes1, const double *modes2 = NULL);
    
		double getMSDQuaternion (const double *Q,const double *T);
        double getMSDQuaternion (const double *Q,const double *T, const double *modes1, const double *modes2 = NULL);
        double getMSDOrthoQuaternion (const double *Q,const double *T, const double *modes1, const double *modes2 = NULL);


		double getMSDTwoTransforms(const double *M1,const double *M2,const double *T1,const double *T2, const double *modes1 = NULL, const double *modes2 = NULL);
        double getMSDTwoTransformsOrtho(const double *M1,const double *M2,const double *T1,const double *T2, const double *modes1 = NULL, const double *modes2 = NULL);
        double getMSDQTwoTransformsOrtho(const double q1[4], const double q2[4],const double *T1,const double *T2, const double *modes1 = NULL, const double *modes2 = NULL);
    
		double getMSDTwoTransformsPAIQuaternion(const double q1[4], const double q2[4], const double T1[3], const double T2[3], const double *modes1 = NULL, const double *modes2 = NULL);
		double getMSDQuaternionTwoMolecules (const double *Q,const double *T);
    
        double getMSDModes(const double *Q,const double *T, const double *modes1, const double *modes2 = NULL);
        double getMSDMatrixModes(const double *M,const double *T, const double *modes1, const double *modes2 = NULL);

        double getMSDOrthoModes(const double *Q,const double *T, const double *modes1, const double *modes2 = NULL);
        double getMSDMatrixOrthoModes(const double *Q,const double *T, const double *modes1, const double *modes2 = NULL);

		//members
		int N;
		double Mass;
};

template <class Vec3>
RMSD<Vec3>::RMSD() {

}

template <class Vec3>
RMSD<Vec3>::~RMSD() {

}

template<class Vec3>
void RMSD<Vec3>::init(const Vec3 *xyz, int N, const double *w){

	computeCOM(xyz, N, w);
	computeInertia(xyz, N, w);
}

template<class Vec3>
void RMSD<Vec3>::initModes(const Vec3 *xyz, int N,  Vec3 **fxyz, int nModes, const double *w){

    init(xyz, N, w);

    nM = nModes;
    computeCOMModes(N, fxyz, w);
    computeCrossCOMModes(xyz, N, fxyz , w) ;
    computeCrossModesModes(N, fxyz , w);
    
}

template<class Vec3>
void RMSD<Vec3>::init(const Vec3 *xyz, const Vec3 *xyz1, int N, const double *w){
	
	computeCOM(xyz, N, w);
	computeInertiaTwoMolecules(xyz, xyz1, N, w);
	computeCrossCoefficients(xyz, xyz1, N, w);
}

//
// RMSD::computeCOM()
// Initialization, performed only once
//
template<class Vec3>
void RMSD<Vec3>::computeCOM(const Vec3 *xyz, int N, const double *w) {

	double x,y,z;
	Mass = 0.0;
	COM[0] = COM[1] = COM[2] = 0.0;

	if (w) {
		for (int i=0; i<N; ++i) {
			Mass+=w[i];

			x=xyz[i][0];
			y=xyz[i][1];
			z=xyz[i][2];


			COM[0]+=x*w[i]; COM[1]+=y*w[i]; COM[2]+=z*w[i];
		} 
	} else {
		for (int i=0; i<N; ++i) {

			x=xyz[i][0];
			y=xyz[i][1];
			z=xyz[i][2];


			COM[0]+=x; COM[1]+=y; COM[2]+=z;
		}
		Mass = N;
	} 
		
	COM[0]/=Mass; COM[1]/=Mass; COM[2]/=Mass;
}

//
// RMSD::computeInertia()
// Initialization, performed only once
//
template <class Vec3>
void RMSD<Vec3>::computeInertia(const Vec3 *xyz, int N, const double *w) {

	double x,y,z;
	Xxx = Xyy = Xzz = 0.0;
	Ixy = Iyz = Izx = 0.0;

	if (w) {
		for (int i=0; i<N; ++i) {

			x=xyz[i][0]-COM[0];
			y=xyz[i][1]-COM[1];
			z=xyz[i][2]-COM[2];

			Xxx+=x*x*w[i]; Xyy+=y*y*w[i]; Xzz+=z*z*w[i];

			Ixy-=x*y*w[i]; Izx-=z*x*w[i]; Iyz-=y*z*w[i];
		}
	} else {
		for (int i=0; i<N; ++i) {

			x=xyz[i][0]-COM[0];
			y=xyz[i][1]-COM[1];
			z=xyz[i][2]-COM[2];

			Xxx+=x*x; Xyy+=y*y; Xzz+=z*z;

			Ixy-=x*y; Izx-=z*x; Iyz-=y*z;
			
		}
	}

	Xxx/=Mass; Xyy/=Mass; Xzz/=Mass;
	Ixy/=Mass; Izx/=Mass; Iyz/=Mass;
	
	Xxx*=4; Xyy*=4; Xzz*=4;
	Ixy*=4; Izx*=4; Iyz*=4;
	
	Ixx = Xyy + Xzz;
	Iyy = Xxx + Xzz;
	Izz = Xxx + Xyy;
	
}

//
// RMSD::computeInertia()
// Initialization, performed only once
//
template <class Vec3>
void RMSD<Vec3>::computeInertiaTwoMolecules(const Vec3 *xyz, const Vec3 *xyz1, int N, const double *w) {
	
	double x,y,z;
	double x1,y1,z1;
	Xxx = Xyy = Xzz = 0.0;
	Ixy = Iyz = Izx = 0.0;
	
	if (w) {
		for (int i=0; i<N; ++i) {
			
			x=xyz[i][0]-COM[0];
			y=xyz[i][1]-COM[1];
			z=xyz[i][2]-COM[2];
			x1=xyz1[i][0]-COM[0];
			y1=xyz1[i][1]-COM[1];
			z1=xyz1[i][2]-COM[2];
			
			Xxx+=x*x1*w[i]; Xyy+=y*y1*w[i]; Xzz+=z*z1*w[i];
			
			Ixy-=(x*y1*w[i] + x1*y*w[i]);
			Izx-=(z*x1*w[i] + z1*x*w[i]);
			Iyz-=(y*z1*w[i] + y1*z*w[i]);
		}
	} else {
		for (int i=0; i<N; ++i) {
			
			x=xyz[i][0]-COM[0];
			y=xyz[i][1]-COM[1];
			z=xyz[i][2]-COM[2];
			x1=xyz1[i][0]-COM[0];
			y1=xyz1[i][1]-COM[1];
			z1=xyz1[i][2]-COM[2];
			
			Xxx+=x*x1; Xyy+=y*y1; Xzz+=z*z1;
			
			Ixy-=(x*y1 + x1*y);
			Izx-=(z*x1 + z1*x);
			Iyz-=(y*z1 + y1*z);
			
		}
	}
	
	Xxx/=Mass; Xyy/=Mass; Xzz/=Mass;
	Ixy/=Mass; Izx/=Mass; Iyz/=Mass;
	
	Xxx*=4; Xyy*=4; Xzz*=4;
	Ixy*=4; Izx*=4; Iyz*=4;
	
	Ixx = Xyy + Xzz;
	Iyy = Xxx + Xzz;
	Izz = Xxx + Xyy;
	
}

//
// RMSD::computeCrossCoefficients()
// Initialization, performed only once
//
template <class Vec3>
void RMSD<Vec3>::computeCrossCoefficients(const Vec3 *xyz, const Vec3 *xyz1, int N, const double *w) {
	
	double x,y,z, x1, y1, z1;
	xy = yz = zx = 0.0;
	xm = ym = zm = 0.0;
	xm2 = ym2 = zm2 = 0.0;
	
	if (w) {
		for (int i=0; i<N; ++i) {
			
			x=xyz[i][0]-COM[0];
			y=xyz[i][1]-COM[1];
			z=xyz[i][2]-COM[2];
			x1=xyz1[i][0]-COM[0];
			y1=xyz1[i][1]-COM[1];
			z1=xyz1[i][2]-COM[2];
			
			xy += w[i]*(y*x1 - y1*x);
			yz += w[i]*(z*y1 - z1*y);
			zx += w[i]*(x*z1 - x1*z);
			
			xm += w[i]*(x - x1);
			ym += w[i]*(y - y1);
			zm += w[i]*(z - z1);
			
			xm2 += w[i]*(x - x1)*(x - x1);
			ym2 += w[i]*(y - y1)*(y - y1);
			zm2 += w[i]*(z - z1)*(z - z1);
		}
	} else {
		for (int i=0; i<N; ++i) {
			
			x=xyz[i][0]-COM[0];
			y=xyz[i][1]-COM[1];
			z=xyz[i][2]-COM[2];
			x1=xyz1[i][0]-COM[0];
			y1=xyz1[i][1]-COM[1];
			z1=xyz1[i][2]-COM[2];
			
			xy += y*x1 - y1*x;
			yz += z*y1 - z1*y;
			zx += x*z1 - x1*z;
			
			xm += x - x1;
			ym += y - y1;
			zm += z - z1;
			
			xm2 += (x - x1)*(x - x1);
			ym2 += (y - y1)*(y - y1);
			zm2 += (z - z1)*(z - z1);
			
		}
	}
	
	xy /= Mass; yz /= Mass; zx /= Mass;
	xm /= Mass; ym /= Mass; zm /= Mass;
	xm2 /= Mass; ym2 /= Mass; zm2 /= Mass;
	
}

//
// RMSD::computeCOModes()
// Initialization, performed only once when normal modes
//
template <class Vec3>
void RMSD<Vec3>::computeCOMModes( int N,  Vec3 **fxyz, const double *w){

    B.resize(nM);
   
    double x,y,z;
    
    for(int j =0; j<nM; ++j) {
        B[j][0] = B[j][1] = B[j][2] = 0.0;
    
        if (w) {
            for (int i=0; i<N; ++i) {
            
                x=fxyz[j][i][0];
                y=fxyz[j][i][1];
                z=fxyz[j][i][2];
                
                
                 B[j][0]+=x*w[i];  B[j][1]+=y*w[i];  B[j][2]+=z*w[i];
            }
        } else {
            for (int i=0; i<N; ++i) {
                x=fxyz[j][i][0];
                y=fxyz[j][i][1];
                z=fxyz[j][i][2];
                
                
                B[j][0]+=x; B[j][1]+=y; B[j][2]+=z;
            }
        }
        //  B[j][0]/=Mass; B[j][1]/=Mass; B[j][2]/=Mass;
    }
}
//
// RMSD::computeCrossCOMModes()
// Initialization, performed only once when normal modes
//
template <class Vec3>
void RMSD<Vec3>::computeCrossCOMModes(const Vec3 *xyz, int N,  Vec3 **fxyz , const double *w) {
    
    double x,y,z;
    double fx,fy,fz;
    
    Dx.resize(nM);
    Dy.resize(nM);
    Dz.resize(nM);

    for (int j =0 ; j < nM ;j++){
        Dx[j][0] = Dx[j][1] = Dx[j][2] = 0.0;
        Dy[j][0] = Dy[j][1] = Dy[j][2] = 0.0;
        Dz[j][0] = Dz[j][1] = Dz[j][2] = 0.0;

        if (w) {
            for (int i=0; i<N; ++i) {
                
                fx=fxyz[j][i][0];
                fy=fxyz[j][i][1];
                fz=fxyz[j][i][2];
                
                x=xyz[i][0];
                y=xyz[i][1];
                z=xyz[i][2];
                
                Dx[j][0]+=x*fx*w[i];  Dx[j][1]+=x*fy*w[i];  Dx[j][2]+=x*fz*w[i];
                Dy[j][0]+=y*fx*w[i];  Dy[j][1]+=y*fy*w[i];  Dy[j][2]+=y*fz*w[i];
                Dz[j][0]+=z*fx*w[i];  Dz[j][1]+=z*fy*w[i];  Dz[j][2]+=z*fz*w[i];

            }
        } else {
            for (int i=0; i<N; ++i) {
                
                fx=fxyz[j][i][0];
                fy=fxyz[j][i][1];
                fz=fxyz[j][i][2];
                
                x=xyz[i][0];
                y=xyz[i][1];
                z=xyz[i][2];
                
                Dx[j][0]+=x*fx;  Dx[j][1]+=x*fy;  Dx[j][2]+=x*fz;
                Dy[j][0]+=y*fx;  Dy[j][1]+=y*fy;  Dy[j][2]+=y*fz;
                Dz[j][0]+=z*fx;  Dz[j][1]+=z*fy;  Dz[j][2]+=z*fz;
            }
        }
    }
}

//
// RMSD::computeCrossModesModes()
// Initialization, performed only once when normal modes
//
template <class Vec3>
void RMSD<Vec3>::computeCrossModesModes(int N,  Vec3 **fxyz , const double *w )
{
    double fjx,fjy,fjz;
    double fkx,fky,fkz;

    Fcol1.resize(nM);
    Fcol2.resize(nM);
    Fcol3.resize(nM);
    
    for (int j =0 ; j < nM ;j++){
        Fcol1[j].resize(nM);
        Fcol2[j].resize(nM);
        Fcol3[j].resize(nM);
        
        for (int k =0 ; k < nM ;k++){

            Fcol1[j][k][0] = Fcol1[j][k][1] = Fcol1[j][k][2] = 0.0;
            Fcol2[j][k][0] = Fcol2[j][k][1] = Fcol2[j][k][2] = 0.0;
            Fcol3[j][k][0] = Fcol3[j][k][1] = Fcol3[j][k][2] = 0.0;
            
            if (w) {
                for (int i=0; i<N; ++i) {
                    
                    fjx=fxyz[j][i][0];
                    fjy=fxyz[j][i][1];
                    fjz=fxyz[j][i][2];
                    fkx=fxyz[k][i][0];
                    fky=fxyz[k][i][1];
                    fkz=fxyz[k][i][2];
                    
                    
                    Fcol1[j][k][0]+=fkx*fjx*w[i];  Fcol1[j][k][1]+=fkx*fjy*w[i];  Fcol1[j][k][2]+=fkx*fjz*w[i];
                    Fcol2[j][k][0]+=fky*fjx*w[i];  Fcol2[j][k][1]+=fky*fjy*w[i];  Fcol2[j][k][2]+=fky*fjz*w[i];
                    Fcol3[j][k][0]+=fkz*fjx*w[i];  Fcol3[j][k][1]+=fkz*fjy*w[i]; Fcol3[j][k][2]+=fkz*fjz*w[i];
                    
                }
            } else {
                for (int i=0; i<N; ++i) {
                    
                    fjx=fxyz[j][i][0];
                    fjy=fxyz[j][i][1];
                    fjz=fxyz[j][i][2];
                    fkx=fxyz[k][i][0];
                    fky=fxyz[k][i][1];
                    fkz=fxyz[k][i][2];
                    
                    
                    Fcol1[j][k][0]+=fkx*fjx;  Fcol1[j][k][1]+=fkx*fjy;  Fcol1[j][k][2]+=fkx*fjz;
                    Fcol2[j][k][0]+=fky*fjx;  Fcol2[j][k][1]+=fky*fjy;  Fcol2[j][k][2]+=fky*fjz;
                    Fcol3[j][k][0]+=fkz*fjx;  Fcol3[j][k][1]+=fkz*fjy;  Fcol3[j][k][2]+=fkz*fjz;
                    
                }
            }
        }
    }
}


//
// RMSD::getMSDQuaternion()
// Total number of operations: 34* and 20+
// Equations 23 and 24
//
template <class Vec3>
double RMSD<Vec3>::getMSDQuaternion(const double q[4], const double TNet[3]) {

	vector3 T;
	double rmsd2=0;

	double cx = q[1]*COM[2]-q[2]*COM[1] + q[3]*COM[0];
	double cy = q[2]*COM[0]-q[0]*COM[2] + q[3]*COM[1];
	double cz = q[0]*COM[1]-q[1]*COM[0] + q[3]*COM[2];

	double dx = 2*(q[1]*cz-q[2]*cy);
	double dy = 2*(q[2]*cx-q[0]*cz);
	double dz = 2*(q[0]*cy-q[1]*cx);

	T[0] = dx + TNet[0];
	T[1] = dy + TNet[1];
	T[2] = dz + TNet[2];	
// number of operations (World frame): 18* and 12+	
// number of operations (COM frame): 0

	// rotational component
	rmsd2=	q[0]*q[1]*Ixy;
	rmsd2+=	q[0]*q[2]*Izx;
	rmsd2+=	q[1]*q[2]*Iyz;
	
	rmsd2 = 2*rmsd2;
	
	rmsd2+=	q[0]*q[0]*Ixx;
	rmsd2+=	q[1]*q[1]*Iyy;
	rmsd2+=	q[2]*q[2]*Izz;
// number of operations (World and COM frames): 13* and 5+
// number of operations (PAI frame): 6* and 2+
	
	// translational component
	rmsd2+=T[0]*T[0]+T[1]*T[1]+T[2]*T[2];
// number of operations: 3* and 3+	

//Total number of operations (World frame): 34* and 20+
//Total number of operations (COM frame): 16* and 8+
//Total number of operations (PAI frame): 9* and 5+

	return rmsd2;
}

//
// RMSD::getMSDQuaternion()
// NMA version
//
template <class Vec3>
double RMSD<Vec3>::getMSDQuaternion(const double q[4], const double TNet[3], const double *modes1, const double *modes2) {
    
    double rmsd2=0;
    rmsd2 += getMSDQuaternion( q, TNet);
    rmsd2 += getMSDModes( q, TNet, modes1, modes2);

    return rmsd2;
}

//
// RMSD::getMSDQuaternion()
// NMA version
//
template <class Vec3>
double RMSD<Vec3>::getMSDOrthoQuaternion(const double q[4], const double TNet[3], const double *modes1, const double *modes2) {

    double rmsd2=0;
    rmsd2 += getMSDQuaternion( q, TNet);
    rmsd2 += getMSDOrthoModes( q, TNet, modes1, modes2);
    return rmsd2;
}


//
// RMSD::getMSDMatrix()
// Total number of operations: 19* and 26+
// Equations 22 and 25
//
template <class Vec3>
double RMSD<Vec3>::getMSDMatrix(const double m[9], const double TNet [3]) {

	double rmsd2=0.0;
	vector3 T;
	
	// T=mat*COM + TNet - COM
	T[0] = (m[0]-1)*COM[0]+m[1]*COM[1]+m[2]*COM[2] + TNet[0];
	T[1] = m[3]*COM[0]+(m[4]-1)*COM[1]+m[5]*COM[2] + TNet[1];
	T[2] = m[6]*COM[0]+m[7]*COM[1]+(m[8]-1)*COM[2] + TNet[2];
// number of operations (World frame): 9* and 12+
// number of operations (COM frame): 0

	//rotational component
	rmsd2 = 	(1-m[0])*Xxx;
	rmsd2 += 	(1-m[4])*Xyy;
	rmsd2 += 	(1-m[8])*Xzz;

	rmsd2 += 	(m[1]+m[3])*Ixy;
	rmsd2 += 	(m[2]+m[6])*Izx;
	rmsd2 += 	(m[5]+m[7])*Iyz;

	rmsd2 =		0.5*rmsd2;

// number of operations (World and COM frames): 7* and 11+
// number of operations (PAI frame): 3* and 5+

	//translational component
	rmsd2 += T[0]*T[0]+T[1]*T[1]+T[2]*T[2];
// number of operations: 3* and 3+	

//Total number of operations (World frame): 19* and 26+
//Total number of operations (COM frame): 10* and 14+
//Total number of operations (PAI frame): 6* and 8+
	return rmsd2;
}

//
// RMSD::getMSDMatrix()
//
template <class Vec3>
double RMSD<Vec3>::getMSDMatrix(const double m[9], const double TNet [3], const double *modes1, const double *modes2) {
   
    double rmsd2=0;
    rmsd2 += getMSDMatrix( m, TNet);
    rmsd2 += getMSDMatrixModes( m, TNet, modes1, modes2);
    
    return rmsd2;
   
}
//
// RMSD::getMSDMatrix()
// NMA version
//
template <class Vec3>
double RMSD<Vec3>::getMSDOrthoMatrix(const double m[9], const double TNet[3], const double *modes1, const double *modes2) {
    
    double rmsd2=0.0;
    rmsd2 += getMSDMatrix( m, TNet);
    rmsd2 += getMSDMatrixOrthoModes( m, TNet, modes1, modes2);
    
    return rmsd2;
}

template <class Vec3>
void RMSD<Vec3>::multiplyMatriceTMatrice(const double *m1, const double *m2, double *m) {
//computes m1.transpose() * m2
    
	m[0] = m1[0]*m2[0] + m1[3]*m2[3] + m1[6]*m2[6];
	m[4] = m1[1]*m2[1] + m1[4]*m2[4] + m1[7]*m2[7];
	m[8] = m1[2]*m2[2] + m1[5]*m2[5] + m1[8]*m2[8];

	m[1] = m1[0]*m2[1] + m1[3]*m2[4] + m1[6]*m2[7];
	m[3] = m2[0]*m1[1] + m2[3]*m1[4] + m2[6]*m1[7];

	m[2] = m1[0]*m2[2] + m1[3]*m2[5] + m1[6]*m2[8];
	m[6] = m2[0]*m1[2] + m2[3]*m1[5] + m2[6]*m1[8];

	m[5] = m1[1]*m2[2] + m1[4]*m2[5] + m1[7]*m2[8];
	m[7] = m2[1]*m1[2] + m2[4]*m1[5] + m2[7]*m1[8];

	// 27* and 18+
}

//
// RMSD::getMSDTwoTransforms()
// Total number of operations: 55* and 59+
// Equation 21
//
template <class Vec3>
double RMSD<Vec3>::getMSDTwoTransforms(const double m1[9], const double m2[9], const double TNet1[3], const double TNet2[3], const double *modes1, const double *modes2 ) {

	double rmsd2=0.0;
	vector3 T1, T2;
	
	// T=mat*COM + TNet - COM
	T1[0] = (m1[0]-1)*COM[0]+m1[1]*COM[1]+m1[2]*COM[2] + TNet1[0];
	T1[1] = m1[3]*COM[0]+(m1[4]-1)*COM[1]+m1[5]*COM[2] + TNet1[1];
	T1[2] = m1[6]*COM[0]+m1[7]*COM[1]+(m1[8]-1)*COM[2] + TNet1[2];

	T2[0] = (m2[0]-1)*COM[0]+m2[1]*COM[1]+m2[2]*COM[2] + TNet2[0];
	T2[1] = m2[3]*COM[0]+(m2[4]-1)*COM[1]+m2[5]*COM[2] + TNet2[1];
	T2[2] = m2[6]*COM[0]+m2[7]*COM[1]+(m2[8]-1)*COM[2] + TNet2[2];
// number of operations (World frame): 18* and 24+
// number of operations (PAI and COM frames): 0

    //    MRelative = M1Tranpose. M2
	double MRelative[9] ;
	multiplyMatriceTMatrice(m1,m2,MRelative);

// number of operations: 27* and 18+
// number of operations (quatornions): 16* and 12+
// or even less! (quatornions): 12* and 9+
	
	//rotational component
	rmsd2 = 	(1-MRelative[0])*Xxx;
	rmsd2 += 	(1-MRelative[4])*Xyy;
	rmsd2 += 	(1-MRelative[8])*Xzz;

	rmsd2 += 	(MRelative[1]+MRelative[3])*Ixy;
	rmsd2 += 	(MRelative[2]+MRelative[6])*Izx;
	rmsd2 += 	(MRelative[5]+MRelative[7])*Iyz;
	
	rmsd2 =		0.5*rmsd2;
// number of operations (Worls and COM frames): 7* and 11+
// number of operations (PAI frame): 6* and 2+

	
	//translational component
	vector3	dT;
	dT[0] =T1[0]-T2[0];
	dT[1] =T1[1]-T2[1];
	dT[2] =T1[2]-T2[2];

	rmsd2 += dT[0]*dT[0] + dT[1]*dT[1] + dT[2]*dT[2];
// number of operations: 3* and 6+

    
// the two transform version is equivalent to the one transform applying m1.transpose to every member.
//    MRelative = M1Tranpose. M2
//    TRelative = M1Transpose.(T2-T1)
    if (modes1) {
        double TRelative[3];

        TRelative[0] = m1[0]*(TNet2[0]-TNet1[0]) + m1[3]*(TNet2[1]-TNet1[1]) + m1[6]*(TNet2[2]-TNet1[2]);
        TRelative[1] = m1[1]*(TNet2[0]-TNet1[0]) + m1[4]*(TNet2[1]-TNet1[1])+ m1[7]*(TNet2[2]-TNet1[2]) ;
        TRelative[2] = m1[2]*(TNet2[0]-TNet1[0]) + m1[5]*(TNet2[1]-TNet1[1]) + m1[8]*(TNet2[2]-TNet1[2]);

        
         rmsd2 += getMSDMatrixModes(MRelative, TRelative, modes1, modes2);
    }
//Total number of operations (World frame, matrix): 55* and 59+ = 114
//Total number of operations (PAI frame, quaternion): 21* and 17+ = 38
	return rmsd2;

}
//
// RMSD::getMSDTwoTransforms()
// Total number of operations: 55* and 59+
// Equation 21
//
template <class Vec3>
double RMSD<Vec3>::getMSDTwoTransformsOrtho(const double m1[9], const double m2[9], const double TNet1[3], const double TNet2[3], const double *modes1, const double *modes2 ) {
    
    double rmsd2=0.0;
    vector3 T1, T2;
    
    // T=mat*COM + TNet - COM
    T1[0] = (m1[0]-1)*COM[0]+m1[1]*COM[1]+m1[2]*COM[2] + TNet1[0];
    T1[1] = m1[3]*COM[0]+(m1[4]-1)*COM[1]+m1[5]*COM[2] + TNet1[1];
    T1[2] = m1[6]*COM[0]+m1[7]*COM[1]+(m1[8]-1)*COM[2] + TNet1[2];
    
    T2[0] = (m2[0]-1)*COM[0]+m2[1]*COM[1]+m2[2]*COM[2] + TNet2[0];
    T2[1] = m2[3]*COM[0]+(m2[4]-1)*COM[1]+m2[5]*COM[2] + TNet2[1];
    T2[2] = m2[6]*COM[0]+m2[7]*COM[1]+(m2[8]-1)*COM[2] + TNet2[2];
    // number of operations (World frame): 18* and 24+
    // number of operations (PAI and COM frames): 0
    
    //    MRelative = M1Tranpose. M2
    double MRelative[9] ;
    multiplyMatriceTMatrice(m1,m2,MRelative);
    
    // number of operations: 27* and 18+
    // number of operations (quatornions): 16* and 12+
    // or even less! (quatornions): 12* and 9+
    
    //rotational component
    rmsd2 = 	(1-MRelative[0])*Xxx;
    rmsd2 += 	(1-MRelative[4])*Xyy;
    rmsd2 += 	(1-MRelative[8])*Xzz;
    
    rmsd2 += 	(MRelative[1]+MRelative[3])*Ixy;
    rmsd2 += 	(MRelative[2]+MRelative[6])*Izx;
    rmsd2 += 	(MRelative[5]+MRelative[7])*Iyz;
    
    rmsd2 =		0.5*rmsd2;
    // number of operations (Worls and COM frames): 7* and 11+
    // number of operations (PAI frame): 6* and 2+
    
    
    //translational component
    vector3	dT;
    dT[0] =T1[0]-T2[0];
    dT[1] =T1[1]-T2[1];
    dT[2] =T1[2]-T2[2];
    
    rmsd2 += dT[0]*dT[0] + dT[1]*dT[1] + dT[2]*dT[2];
    // number of operations: 3* and 6+
    
    
    // the two transform version is equivalent to the one transform applying m1.transpose to every member.
    //    MRelative = M1Tranpose. M2
    //    TRelative = M1Transpose.(T2-T1)
    if (modes1) {
        double TRelative[3];
        
        TRelative[0] = m1[0]*(TNet2[0]-TNet1[0]) + m1[3]*(TNet2[1]-TNet1[1]) + m1[6]*(TNet2[2]-TNet1[2]);
        TRelative[1] = m1[1]*(TNet2[0]-TNet1[0]) + m1[4]*(TNet2[1]-TNet1[1])+ m1[7]*(TNet2[2]-TNet1[2]) ;
        TRelative[2] = m1[2]*(TNet2[0]-TNet1[0]) + m1[5]*(TNet2[1]-TNet1[1]) + m1[8]*(TNet2[2]-TNet1[2]);
        
        
        rmsd2 += getMSDMatrixOrthoModes(MRelative, TRelative, modes1, modes2);
    }
    //Total number of operations (World frame, matrix): 55* and 59+ = 114
    //Total number of operations (PAI frame, quaternion): 21* and 17+ = 38
    return rmsd2;
    
}

//
// RMSD::getMSDTwoTransforms()
// Total number of operations: 55* and 59+
// Equation 21
//
template <class Vec3>
double RMSD<Vec3>::getMSDQTwoTransformsOrtho(const double q1[4], const double q2[4], const double TNet1[3], const double TNet2[3], const double *modes1, const double *modes2 ) {
    
    double rmsd2=0.0;
    vector3 T1, T2;
    
    // T=mat*COM + TNet - COM
    
    double cx = q1[1]*COM[2]-q1[2]*COM[1] + q1[3]*COM[0];
    double cy = q1[2]*COM[0]-q1[0]*COM[2] + q1[3]*COM[1];
    double cz = q1[0]*COM[1]-q1[1]*COM[0] + q1[3]*COM[2];
    
    double dx = 2*(q1[1]*cz-q1[2]*cy);
    double dy = 2*(q1[2]*cx-q1[0]*cz);
    double dz = 2*(q1[0]*cy-q1[1]*cx);
    
    T1[0] = dx + TNet1[0];
    T1[1] = dy + TNet1[1];
    T1[2] = dz + TNet1[2];
    
    
     cx = q2[1]*COM[2]-q2[2]*COM[1] + q2[3]*COM[0];
     cy = q2[2]*COM[0]-q2[0]*COM[2] + q2[3]*COM[1];
     cz = q2[0]*COM[1]-q2[1]*COM[0] + q2[3]*COM[2];
    
     dx = 2*(q2[1]*cz-q2[2]*cy);
     dy = 2*(q2[2]*cx-q2[0]*cz);
     dz = 2*(q2[0]*cy-q2[1]*cx);
    
    T2[0] = dx + TNet2[0];
    T2[1] = dy + TNet2[1];
    T2[2] = dz + TNet2[2];
    

    // number of operations (World frame): 18* and 24+
    // number of operations (PAI and COM frames): 0
    
    
    double q[3] ;
    // multiplication of inverse q1 and q2
    q[0] = q1[3]*q2[0] - q1[0]*q2[3]  -  q1[1]*q2[2]  +  q1[2]*q2[1];
    q[1] = q1[3]*q2[1] - q1[1]*q2[3]  -  q1[2]*q2[0]  +  q1[0]*q2[2];
    q[2] = q1[3]*q2[2] - q1[2]*q2[3]  -  q1[0]*q2[1]  +  q1[1]*q2[0];
 
    // number of operations: 27* and 18+
    // number of operations (quatornions): 16* and 12+
    // or even less! (quatornions): 12* and 9+


    // rotational component
    rmsd2=	q[0]*q[1]*Ixy;
    rmsd2+=	q[0]*q[2]*Izx;
    rmsd2+=	q[1]*q[2]*Iyz;
    
    rmsd2 = 2*rmsd2;
    
    rmsd2+=	q[0]*q[0]*Ixx;
    rmsd2+=	q[1]*q[1]*Iyy;
    rmsd2+=	q[2]*q[2]*Izz;
  
    // number of operations (World and COM frames): 13* and 5+
    // number of operations (PAI frame): 6* and 2+
    
    //translational component
    vector3	dT;
    dT[0] =T1[0]-T2[0];
    dT[1] =T1[1]-T2[1];
    dT[2] =T1[2]-T2[2];
    
    rmsd2 += dT[0]*dT[0] + dT[1]*dT[1] + dT[2]*dT[2];
    // number of operations: 3* and 6+
    
    
    // the two transform version is equivalent to the one transform applying m1.transpose to every member.
    //    MRelative = M1Tranpose. M2
    //    TRelative = M1Transpose.(T2-T1)
    if (modes1) {
        double TRelative[3];
        
         cx = q[1]*(TNet2[2]-TNet1[2])-q[2]*(TNet2[1]-TNet1[1]) + q[3]*(TNet2[0]-TNet1[0]);
         cy = q[2]*(TNet2[0]-TNet1[0])-q[0]*(TNet2[2]-TNet1[2]) + q[3]*(TNet2[1]-TNet1[1]);
         cz = q[0]*(TNet2[1]-TNet1[1])-q[1]*(TNet2[0]-TNet1[0]) + q[3]*(TNet2[2]-TNet1[2]);
        
         TRelative[0] = 2*(q[1]*cz-q[2]*cy);
         TRelative[1] = 2*(q[2]*cx-q[0]*cz);
         TRelative[2] = 2*(q[0]*cy-q[1]*cx);
        
        rmsd2 += getMSDOrthoQuaternion(q, TRelative, modes1, modes2);
    }
    //Total number of operations (World frame, matrix): 55* and 59+ = 114
    //Total number of operations (PAI frame, quaternion): 21* and 17+ = 38
    return rmsd2;
    
}


//
// RMSD::getMSDTwoTransformsPAIQuaternion()
// Total number of operations: 21* and 17+
// Equation 26
//
template <class Vec3>
double RMSD<Vec3>::getMSDTwoTransformsPAIQuaternion(const double q1[4], const double q2[4], const double T1[3], const double T2[3], const double *modes1 , const double *modes2 ) {

	double rmsd2=0.0;

	
	double q[3] ;
	// multiplication of inverse q1 and q2
	q[0] = q1[3]*q2[0] - q1[0]*q2[3]  -  q1[1]*q2[2]  +  q1[2]*q2[1];
	q[1] = q1[3]*q2[1] - q1[1]*q2[3]  -  q1[2]*q2[0]  +  q1[0]*q2[2];
	q[2] = q1[3]*q2[2] - q1[2]*q2[3]  -  q1[0]*q2[1]  +  q1[1]*q2[0];

	// number of operations (quatornions): 12* and 9+
	
	//rotational component
	rmsd2=	q[0]*q[0]*Ixx;
	rmsd2+=	q[1]*q[1]*Iyy;
	rmsd2+=	q[2]*q[2]*Izz;
	// number of operations (PAI frame): 6* and 2+

	//translational component
	vector3	dT;
	dT[0] =T1[0]-T2[0];
	dT[1] =T1[1]-T2[1];
	dT[2] =T1[2]-T2[2];
	
	rmsd2 += dT[0]*dT[0] + dT[1]*dT[1] + dT[2]*dT[2];
    
	// number of operations: 3* and 6+
    
    // the two transform version is equivalent to the one transform applying m1.transpose to every member.
    //    MRelative = M1Tranpose. M2
    //    TRelative = M1Transpose.(T2-T1)

    // quaternion gives [s,q] =
    if (modes1) {
        
        double QRelative[4];
        QRelative[0]=q[0];  QRelative[1]=q[1];  QRelative[2]=q[2];
        QRelative[3] = q1[3]*q2[3] + q1[0]*q2[0] + q1[1]*q2[1] + q1[2]*q2[2];
        
        double TRelative[3];
        
        double m1[9];
        double a = q1[3];
        double b = q1[0];
        double c = q1[1];
        double d = q1[2];
        
        double a2 = a*a;
        double b2 = b*b;
        double c2 = c*c;
        double d2 = d*d;
        
        m1[0] = a2+b2-c2-d2; m1[4] = a2-b2+c2-d2; m1[8] = a2-b2-c2+d2;
        
        
        double rxy = 2*b*c;
        double rxz = 2*b*d;
        double ryz = 2*c*d;
        double rsx = 2*a*b;
        double rsy = 2*a*c;
        double rsz = 2*a*d;
        
        m1[3] = rxy + rsz;
        m1[1] = rxy - rsz;
        m1[6] = rxz - rsy;
        m1[2] = rxz + rsy;
        m1[7] = ryz + rsx;
        m1[5] = ryz - rsx;
        
        
        TRelative[0] = m1[0]*(T2[0]-T1[0]) + m1[3]*(T2[1]-T1[1]) + m1[6]*(T2[2]-T1[2]);
        TRelative[1] = m1[1]*(T2[0]-T1[0]) + m1[4]*(T2[1]-T1[1])+ m1[7]*(T2[2]-T1[2]) ;
        TRelative[2] = m1[2]*(T2[0]-T1[0]) + m1[5]*(T2[1]-T1[1]) + m1[8]*(T2[2]-T1[2]);
        
        
        rmsd2 += getMSDQuaternion(q, TRelative, modes1, modes2);
    }
	//Total number of operations (PAI frame, quaternion): 21* and 17+ = 38
	return rmsd2;
	
}

//
// RMSD::getMSDQuaternionTwoMolecules()
// Total number of operations: 34* and 20+
// Equations 23 and 24
//
template <class Vec3>
double RMSD<Vec3>::getMSDQuaternionTwoMolecules(const double q[4], const double TNet[3]) {
	
	vector3 T;
	double rmsd2=0;
	
	double cx = q[1]*COM[2]-q[2]*COM[1] + q[3]*COM[0];
	double cy = q[2]*COM[0]-q[0]*COM[2] + q[3]*COM[1];
	double cz = q[0]*COM[1]-q[1]*COM[0] + q[3]*COM[2];
	
	double dx = 2*(q[1]*cz-q[2]*cy);
	double dy = 2*(q[2]*cx-q[0]*cz);
	double dz = 2*(q[0]*cy-q[1]*cx);
	
	T[0] = dx + TNet[0];
	T[1] = dy + TNet[1];
	T[2] = dz + TNet[2];
	// number of operations (World frame): 18* and 12+
	// number of operations (COM frame): 0
	
	// rotational component
	rmsd2=	q[0]*q[1]*Ixy;
	rmsd2+=	q[0]*q[2]*Izx;
	rmsd2+=	q[1]*q[2]*Iyz;
		
	rmsd2+=	q[0]*q[0]*Ixx;
	rmsd2+=	q[1]*q[1]*Iyy;
	rmsd2+=	q[2]*q[2]*Izz;
	// number of operations (World and COM frames): 13* and 5+
	// number of operations (PAI frame): 6* and 2+
	
	// translational component
	rmsd2+=T[0]*T[0]+T[1]*T[1]+T[2]*T[2];
	// number of operations: 3* and 3+
	
	// translational component
	rmsd2+= xm2 + ym2 + zm2;
	rmsd2+= 2*T[0]*xm + 2*T[1]*ym + 2*T[2]*zm;
	// number of operations:
	
	// finally, the last term:
	rmsd2+= 4*q[3]*( q[0]*yz + q[1]*zx + q[2]*xy);
	
	//Total number of operations (World frame):
	//Total number of operations (COM frame):
	//Total number of operations (PAI frame): 
	return rmsd2;
}

//
// RMSD::getMSDModes()
//
//
template <class Vec3>
double RMSD<Vec3>::getMSDModes(const double Q[4],const double T[3], const double *modes1, const double *modes2){
    
    double rmsd2 = 0.0;
    double TB1x = 0.0;
    double TB1y = 0.0;
    double TB1z = 0.0;
    double TB2x = 0.0;
    double TB2y = 0.0;
    double TB2z = 0.0;

    double m[9];
    double a = Q[3];
    double b = Q[0];
    double c = Q[1];
    double d = Q[2];
    
    double a2 = a*a;
    double b2 = b*b;
    double c2 = c*c;
    double d2 = d*d;

    m[0] = a2+b2-c2-d2; m[4] = a2-b2+c2-d2; m[8] = a2-b2-c2+d2;


    double rxy = 2*b*c;
    double rxz = 2*b*d;
    double ryz = 2*c*d;
    double rsx = 2*a*b;
    double rsy = 2*a*c;
    double rsz = 2*a*d;

    m[3] = rxy + rsz;
    m[1] = rxy - rsz;
    m[6] = rxz - rsy;
    m[2] = rxz + rsy;
    m[7] = ryz + rsx;
    m[5] = ryz - rsx;
    
   // std::cout<<" R \n"<<R[0][0] <<" "<<R[0][1] <<" "<<R[0][2] <<"\n "<<R[1][0] <<" " <<R[1][1] <<" "<<R[1][2] <<" \n"<<R[2][0] <<" "<<R[2][1] <<" "<<R[2][2] <<" \n"; // RMSD function call
    for (int j = 0; j< nM; ++j){
        
        ////orthonormal modes
    //    rmsd2+= (modes1[j]*modes1[j] + modes2[j]*modes2[j]);
        
        //translational component modes1 - part1
        TB1x+= modes1[j]*B[j][0];
        TB1y+= modes1[j]*B[j][1];
        TB1z+= modes1[j]*B[j][2];

        //rotational component modes1
        rmsd2 -= 2*modes1[j]*( Dx[j][0]*(m[0]-1) + Dx[j][1]*m[3] + Dx[j][2]*m[6]) ;
        rmsd2 -= 2*modes1[j]*( Dy[j][0]*m[1] + Dy[j][1]*(m[4]-1) + Dy[j][2]*m[7]) ;
        rmsd2 -= 2*modes1[j]*( Dz[j][0]*m[2] + Dz[j][1]*m[5] + Dz[j][2]*(m[8]-1)) ;
        
        if (modes2) {
            //translational component modes2 - part1
            TB2x+= modes2[j]*B[j][0];
            TB2y+= modes2[j]*B[j][1];
            TB2z+= modes2[j]*B[j][2];
            
            //rotational component modes
            rmsd2 -= 2*modes2[j]*( Dx[j][0]*(m[0]-1) + Dx[j][1]*m[1] + Dx[j][2]*m[2]) ;
            rmsd2 -= 2*modes2[j]*( Dy[j][0]*m[3] + Dy[j][1]*(m[4]-1) + Dy[j][2]*m[5]) ;
            rmsd2 -= 2*modes2[j]*( Dz[j][0]*m[6] + Dz[j][1]*m[7] + Dz[j][2]*(m[8]-1)) ;
        }
        
        for (int k = 0; k< nM; ++k){
            if (modes2) {
                // cross modes terms
                rmsd2 -= 2*modes2[j]*modes1[k]*( Fcol1[j][k][0]*m[0] + Fcol1[j][k][1]*m[1] + Fcol1[j][k][2]*m[2]) ;
                rmsd2 -= 2*modes2[j]*modes1[k]*( Fcol2[j][k][0]*m[3] + Fcol2[j][k][1]*m[4] + Fcol2[j][k][2]*m[5]) ;
                rmsd2 -= 2*modes2[j]*modes1[k]*( Fcol3[j][k][0]*m[6] + Fcol3[j][k][1]*m[7] + Fcol3[j][k][2]*m[8]) ;
                //quadratic modes terms
                rmsd2+= (modes2[j]*modes2[k])* ( Fcol1[j][k][0]+ Fcol2[j][k][1] +Fcol3[j][k][2]);
            }
            
            //quadratic modes terms
            rmsd2+= (modes1[j]*modes1[k])* ( Fcol1[j][k][0]+ Fcol2[j][k][1] +Fcol3[j][k][2]);
            
        }
        
    }

    //translational component modes1 -part2
    rmsd2-= 2*T[0]*TB1x + 2*T[1]*TB1y + 2*T[2]*TB1z;
    
    //translational component modes2 - part2
    double bx = T[0]*m[0] + T[1]*m[3] + T[2]*m[6] ;
    
    double by = T[0]*m[1] + T[1]*m[4] + T[2]*m[7] ;
    
    double bz = T[0]*m[2] + T[1]*m[5] + T[2]*m[8] ;
    

    rmsd2+= 2*bx*TB2x + 2*by*TB2y + 2*bz*TB2z;

    
    rmsd2 /=Mass;

    return rmsd2;
    
}
//
// RMSD::getMSDMatrixModes()
//
//
template <class Vec3>
double RMSD<Vec3>::getMSDMatrixModes(const double m[9],const double T[3], const double *modes1, const double *modes2){
    
    double rmsd2 = 0.0;
    double TB1x = 0.0;
    double TB1y = 0.0;
    double TB1z = 0.0;
    double TB2x = 0.0;
    double TB2y = 0.0;
    double TB2z = 0.0;
    

    for (int j = 0; j< nM; ++j){
        
        ////orthonormal modes
        //    rmsd2+= (modes1[j]*modes1[j] + modes2[j]*modes2[j]);
        
        //translational component modes1 - part1
        TB1x+= modes1[j]*B[j][0];
        TB1y+= modes1[j]*B[j][1];
        TB1z+= modes1[j]*B[j][2];
        
        //rotational component modes1
        rmsd2 -= 2*modes1[j]*( Dx[j][0]*(m[0]-1) + Dx[j][1]*m[3] + Dx[j][2]*m[6]) ;
        rmsd2 -= 2*modes1[j]*( Dy[j][0]*m[1] + Dy[j][1]*(m[4]-1) + Dy[j][2]*m[7]) ;
        rmsd2 -= 2*modes1[j]*( Dz[j][0]*m[2] + Dz[j][1]*m[5] + Dz[j][2]*(m[8]-1)) ;
        
        if (modes2) {
            //translational component modes2 - part1
            TB2x+= modes2[j]*B[j][0];
            TB2y+= modes2[j]*B[j][1];
            TB2z+= modes2[j]*B[j][2];
            
            //rotational component modes
            rmsd2 -= 2*modes2[j]*( Dx[j][0]*(m[0]-1) + Dx[j][1]*m[1] + Dx[j][2]*m[2]) ;
            rmsd2 -= 2*modes2[j]*( Dy[j][0]*m[3] + Dy[j][1]*(m[4]-1) + Dy[j][2]*m[5]) ;
            rmsd2 -= 2*modes2[j]*( Dz[j][0]*m[6] + Dz[j][1]*m[7] + Dz[j][2]*(m[8]-1)) ;
        }
        
        for (int k = 0; k< nM; ++k){
            if (modes2) {
                // cross modes terms
                rmsd2 -= 2*modes2[j]*modes1[k]*( Fcol1[j][k][0]*m[0] + Fcol1[j][k][1]*m[1] + Fcol1[j][k][2]*m[2]) ;
                rmsd2 -= 2*modes2[j]*modes1[k]*( Fcol2[j][k][0]*m[3] + Fcol2[j][k][1]*m[4] + Fcol2[j][k][2]*m[5]) ;
                rmsd2 -= 2*modes2[j]*modes1[k]*( Fcol3[j][k][0]*m[6] + Fcol3[j][k][1]*m[7] + Fcol3[j][k][2]*m[8]) ;
                //quadratic modes terms
                rmsd2+= (modes2[j]*modes2[k])* ( Fcol1[j][k][0]+ Fcol2[j][k][1] +Fcol3[j][k][2]);
            }
            
            //quadratic modes terms
            rmsd2+= (modes1[j]*modes1[k])* ( Fcol1[j][k][0]+ Fcol2[j][k][1] +Fcol3[j][k][2]);
            
        }
        
    }
    
    //translational component modes1 -part2
    rmsd2-= 2*T[0]*TB1x + 2*T[1]*TB1y + 2*T[2]*TB1z;
    
    //translational component modes2 - part2
    double bx = T[0]*m[0] + T[1]*m[3] + T[2]*m[6] ;
    
    double by = T[0]*m[1] + T[1]*m[4] + T[2]*m[7] ;
    
    double bz = T[0]*m[2] + T[1]*m[5] + T[2]*m[8] ;
    
    
    rmsd2+= 2*bx*TB2x + 2*by*TB2y + 2*bz*TB2z;
    
    
    rmsd2 /=Mass;
    
    return rmsd2;
    
}
    
//
// RMSD::getMSDModes()
//
// orthonormal modes version
template <class Vec3>
double RMSD<Vec3>::getMSDOrthoModes(const double Q[4],const double T[3], const double *modes1, const double *modes2)
    {
    

    double rmsd2 = 0.0;
    double TB1x = 0.0;
    double TB1y = 0.0;
    double TB1z = 0.0;
    double TB2x = 0.0;
    double TB2y = 0.0;
    double TB2z = 0.0;

        double m[9];
        double a = Q[3];
        double b = Q[0];
        double c = Q[1];
        double d = Q[2];
        
        double a2 = a*a;
        double b2 = b*b;
        double c2 = c*c;
        double d2 = d*d;
        
        m[0] = a2+b2-c2-d2; m[4] = a2-b2+c2-d2; m[8] = a2-b2-c2+d2;
        
        
        double rxy = 2*b*c;
        double rxz = 2*b*d;
        double ryz = 2*c*d;
        double rsx = 2*a*b;
        double rsy = 2*a*c;
        double rsz = 2*a*d;
        
        m[3] = rxy + rsz;
        m[1] = rxy - rsz;
        m[6] = rxz - rsy;
        m[2] = rxz + rsy;
        m[7] = ryz + rsx;
        m[5] = ryz - rsx;

    for (int j = 0; j< nM; ++j){

        //translational component modes1 - part1
        TB1x+= modes1[j]*B[j][0];
        TB1y+= modes1[j]*B[j][1];
        TB1z+= modes1[j]*B[j][2];

        //rotational component modes1
        rmsd2 -= 2*modes1[j]*( Dx[j][0]*(m[0]-1) + Dx[j][1]*m[3] + Dx[j][2]*m[6]) ;
        rmsd2 -= 2*modes1[j]*( Dy[j][0]*m[1] + Dy[j][1]*(m[4]-1) + Dy[j][2]*m[7]) ;
        rmsd2 -= 2*modes1[j]*( Dz[j][0]*m[2] + Dz[j][1]*m[5] + Dz[j][2]*(m[8]-1)) ;

        if (modes2) {
            //translational component modes2 - part1
            TB2x+= modes2[j]*B[j][0];
            TB2y+= modes2[j]*B[j][1];
            TB2z+= modes2[j]*B[j][2];

            //rotational component modes2
            rmsd2 -= 2*modes2[j]*( Dx[j][0]*(m[0]-1) + Dx[j][1]*m[1] + Dx[j][2]*m[2]) ;
            rmsd2 -= 2*modes2[j]*( Dy[j][0]*m[3] + Dy[j][1]*(m[4]-1) + Dy[j][2]*m[5]) ;
            rmsd2 -= 2*modes2[j]*( Dz[j][0]*m[6] + Dz[j][1]*m[7] + Dz[j][2]*(m[8]-1)) ;
        }

        ////orthonormal modes
        //quadratic modes terms
        rmsd2+= (modes1[j]*modes1[j]);

        if (modes2) {
            //orthonormal modes
           rmsd2+= (modes2[j]*modes2[j]);

            // cross modes terms
            for (int k = 0; k< nM; ++k){

                rmsd2 -= 2*modes2[j]*modes1[k]*( Fcol1[j][k][0]*m[0] + Fcol1[j][k][1]*m[1] + Fcol1[j][k][2]*m[2]) ;
                rmsd2 -= 2*modes2[j]*modes1[k]*( Fcol2[j][k][0]*m[3] + Fcol2[j][k][1]*m[4] + Fcol2[j][k][2]*m[5]) ;
                rmsd2 -= 2*modes2[j]*modes1[k]*( Fcol3[j][k][0]*m[6] + Fcol3[j][k][1]*m[7] + Fcol3[j][k][2]*m[8]) ;
            }
        }

    }

    //translational component modes1 -part2
    rmsd2-= 2*T[0]*TB1x + 2*T[1]*TB1y + 2*T[2]*TB1z;

    //translational component modes2 - part2
    double bx = T[0]*m[0] + T[1]*m[3] + T[2]*m[6] ;
    
    double by = T[0]*m[1] + T[1]*m[4] + T[2]*m[7] ;
    
    double bz = T[0]*m[2] + T[1]*m[5] + T[2]*m[8] ;

    rmsd2+= 2*bx*TB2x + 2*by*TB2y + 2*bz*TB2z;
    
    
    rmsd2 /=Mass;
    
    return rmsd2;
    
}


//
// RMSD::getMSDModes()
// matrix + orthonormal modes version
//
template <class Vec3>
double RMSD<Vec3>::getMSDMatrixOrthoModes(const double m[9],const double T[3], const double *modes1, const double *modes2){
    
    double rmsd2 = 0.0;
    double TB1x = 0.0;
    double TB1y = 0.0;
    double TB1z = 0.0;
    double TB2x = 0.0;
    double TB2y = 0.0;
    double TB2z = 0.0;
    
    
    
    for (int j = 0; j< nM; ++j){
        
        //translational component modes1 - part1
        TB1x+= modes1[j]*B[j][0];
        TB1y+= modes1[j]*B[j][1];
        TB1z+= modes1[j]*B[j][2];
        
        
        //rotational component modes1
        rmsd2 -= 2*modes1[j]*( Dx[j][0]*(m[0]-1) + Dx[j][1]*m[3] + Dx[j][2]*m[6]) ;
        rmsd2 -= 2*modes1[j]*( Dy[j][0]*m[1] + Dy[j][1]*(m[4]-1) + Dy[j][2]*m[7]) ;
        rmsd2 -= 2*modes1[j]*( Dz[j][0]*m[2] + Dz[j][1]*m[5] + Dz[j][2]*(m[8]-1)) ;
        
        if (modes2) {
            //translational component modes2 - part1
            TB2x+= modes2[j]*B[j][0];
            TB2y+= modes2[j]*B[j][1];
            TB2z+= modes2[j]*B[j][2];
            
            //rotational component modes2
            rmsd2 -= 2*modes2[j]*( Dx[j][0]*(m[0]-1) + Dx[j][1]*m[1] + Dx[j][2]*m[2]) ;
            rmsd2 -= 2*modes2[j]*( Dy[j][0]*m[3] + Dy[j][1]*(m[4]-1) + Dy[j][2]*m[5]) ;
            rmsd2 -= 2*modes2[j]*( Dz[j][0]*m[6] + Dz[j][1]*m[7] + Dz[j][2]*(m[8]-1)) ;
        }
        
        ////orthonormal modes
        //quadratic modes terms
        rmsd2+= (modes1[j]*modes1[j]);
        
        if (modes2) {
            //orthonormal modes
            rmsd2+= (modes2[j]*modes2[j]);
            
            // cross modes terms
            for (int k = 0; k< nM; ++k){
                
                rmsd2 -= 2*modes2[j]*modes1[k]*( Fcol1[j][k][0]*m[0] + Fcol1[j][k][1]*m[1] + Fcol1[j][k][2]*m[2]) ;
                rmsd2 -= 2*modes2[j]*modes1[k]*( Fcol2[j][k][0]*m[3] + Fcol2[j][k][1]*m[4] + Fcol2[j][k][2]*m[5]) ;
                rmsd2 -= 2*modes2[j]*modes1[k]*( Fcol3[j][k][0]*m[6] + Fcol3[j][k][1]*m[7] + Fcol3[j][k][2]*m[8]) ;
                
            }
        }
    }
    
    //translational component modes1 -part2
    rmsd2-= 2*T[0]*TB1x + 2*T[1]*TB1y + 2*T[2]*TB1z;
    
    //translational component modes2 - part2
    double bx = T[0]*m[0] + T[1]*m[3] + T[2]*m[6] ;
    
    double by = T[0]*m[1] + T[1]*m[4] + T[2]*m[7] ;
    
    double bz = T[0]*m[2] + T[1]*m[5] + T[2]*m[8] ;
    
    rmsd2+= 2*bx*TB2x + 2*by*TB2y + 2*bz*TB2z;
    
    
    rmsd2 /=Mass;
    
    return rmsd2;
    
}

