//Copyright INRIA/CNRS
//
//Neveu, Emilie; Popov, Petr; Hoffmann, Alexandre; Grudinin, Sergei. 2017
//sergei.grudinin@inria.fr
//
//This software is a computer program whose purpose is to compute
//the root mean sum of squared deviations (RMSDs) corresponding
//to a rigid-body motion in constatnt time.
//
//This software is governed by the CeCILL-B  license (fully BSD-compatible)
//under French law and abiding by the rules of distribution of free software.
//You can  use, modify and/ or redistribute the software under the terms
//of the CeCILL-B license as circulated by CEA, CNRS and INRIA at the
//following URL "http://www.cecill.info".
//
//As a counterpart to the access to the source code and  rights to copy,
//modify and redistribute granted by the license, users are provided only
//with a limited warranty  and the software's author,  the holder of the
//economic rights,  and the successive licensors  have only  limited
//liability.
//
//In this respect, the user's attention is drawn to the risks associated
//with loading,  using,  modifying and/or developing or reproducing the
//software by the user in light of its specific status of free software,
//that may mean  that it is complicated to manipulate,  and  that  also
//therefore means  that it is reserved for developers  and  experienced
//professionals having in-depth computer knowledge. Users are therefore
//encouraged to load and test the software's suitability as regards their
//requirements in conditions enabling the security of their systems and/or
//data to be ensured and,  more generally, to use and operate it in the
//same conditions as regards security.
//
//The fact that you are presently reading this means that you have had
//knowledge of the CeCILL-B license and that you accept its terms.

/*
 *=====================================================
 *== example of clustering with collective motions ====
 *=====================================================
 */

#include "RapidRMSD.hpp" // updated library
#include <stdlib.h>
#include <vector>
#include "cPDBAtom.hpp" // for the PDB input
//#include <math.h> // for sqrt
#include <time.h> // for srand

#include <iostream>
#include <fstream>
#include <sstream>
#include <string> // this should be already included in <sstream>

// for the timer to evaluate elapsed time.
#if defined(_WIN32)
#include <time.h>
#include <Windows.h>

#if defined(_MSC_VER) || defined(_MSC_EXTENSIONS)
    #define DELTA_EPOCH_IN_MICROSECS  11644473600000000Ui64
#else
    #define DELTA_EPOCH_IN_MICROSECS  11644473600000000ULL
#endif

struct timezone
{
    int  tz_minuteswest; /* minutes W of Greenwich */
    int  tz_dsttime;     /* type of dst correction */
};

int gettimeofday(struct timeval *tv, struct timezone *tz)
{
    FILETIME ft;
    unsigned __int64 tmpres = 0;
    static int tzflag;

    if (NULL != tv)
        {
        GetSystemTimeAsFileTime(&ft);

        tmpres |= ft.dwHighDateTime;
        tmpres <<= 32;
        tmpres |= ft.dwLowDateTime;

        /*converting file time to unix epoch*/
        tmpres -= DELTA_EPOCH_IN_MICROSECS;
        tmpres /= 10;  /*convert into microseconds*/
        tv->tv_sec = (long)(tmpres / 1000000UL);
        tv->tv_usec = (long)(tmpres % 1000000UL);
        }

    if (NULL != tz)
        {
        if (!tzflag)
            {
            _tzset();
            tzflag++;
            }
        tz->tz_minuteswest = _timezone / 60;
        tz->tz_dsttime = _daylight;
        }
    return 0;
}
#elif defined(__unix__) || ( defined(__APPLE__) && defined (__MACH__) )
    #include <sys/time.h>	// gettimeofday is already there
    #include <cmath>		// for sqrt
#endif


typedef double vector3[3];

typedef double quat[4]; // qx, qy, qz, s

struct timer {

    struct timeval _start, _end;
    double mtime, seconds, useconds;

    void start() {
        gettimeofday(&_start, NULL);
    }

    double stop() {
        gettimeofday(&_end, NULL);
        seconds  = _end.tv_sec  - _start.tv_sec;
        useconds = _end.tv_usec - _start.tv_usec;
        mtime = ((seconds) * 1000.0 + useconds/1000.0) + 0.5;
        printf("Elapsed time: %12lf seconds\n", double (mtime)*0.001);
        return double (mtime)*0.001;
    }

    double stop(const char *c) {
        printf("\n%s :\n",c);
        return stop();
    }

};


void computeMatrix (double* m, double *q) {
    
    if (m && q) {
        double a = q[3];
        double b = q[0];
        double c = q[1];
        double d = q[2];
        
        m[0] = a*a+b*b-c*c-d*d; m[4] = a*a-b*b+c*c-d*d; m[8] = a*a-b*b-c*c+d*d;
        m[1] = 2*b*c-2*a*d; m[3] = 2*b*c+2*a*d;
        m[2] = 2*b*d+2*a*c; m[6] = 2*b*d-2*a*c;
        m[5] = 2*c*d-2*a*b; m[7] = 2*c*d+2*a*b;
        
        return;
    }
    else {
        std::cerr << "Error"<<std::endl;
        exit(0);
    }
}

void mulQuat(quat &r, const quat &a, const quat &b) {

    r[3] = a[3]*b[3] - a[0]*b[0]  -  a[1]*b[1]  -  a[2]*b[2];
    r[0] = a[3]*b[0] + a[0]*b[3]  +  a[1]*b[2]  -  a[2]*b[1];
    r[1] = a[3]*b[1] + a[1]*b[3]  +  a[2]*b[0]  -  a[0]*b[2];
    r[2] = a[3]*b[2] + a[2]*b[3]  +  a[0]*b[1]  -  a[1]*b[0];

    return;
}

void inverseQuat(quat &inverse, const quat &q) {
    inverse[0] = - q[0];
    inverse[1] = - q[1];
    inverse[2] = - q[2];
    inverse[3] = q[3];
}

void relativeRigidTransformation (quat &relativeQ12,  vector3 &relativeT12, const quat &q1, const vector3 T1,
                                 const quat &q2, const vector3 &T2) {

    quat inverseQ1;
    // relative rotation
    inverseQuat(inverseQ1, q1);
    mulQuat(relativeQ12, q2, inverseQ1);  // q2.q1.inv()

    // relative translation : compute T2 -  relativerotation times T1

    
    double A = relativeQ12[3]*relativeQ12[3]-(relativeQ12[0]*relativeQ12[0]+relativeQ12[1]*relativeQ12[1]+relativeQ12[2]*relativeQ12[2]);
    double B = (T1[0]*relativeQ12[0]+T1[1]*relativeQ12[1]+T1[2]*relativeQ12[2]);
    double  cx = relativeQ12[1]*T1[2]-relativeQ12[2]*T1[1];
    double cy = relativeQ12[2]*T1[0]-relativeQ12[0]*T1[2];
    double cz = relativeQ12[0]*T1[1]-relativeQ12[1]*T1[0];
    double xnew = A*T1[0]+2*relativeQ12[3]*cx+2*B*relativeQ12[0];
    double ynew = A*T1[1]+2*relativeQ12[3]*cy+2*B*relativeQ12[1];
    double znew = A*T1[2]+2*relativeQ12[3]*cz+2*B*relativeQ12[2];

    relativeT12[0] = T2[0] - xnew ;
    relativeT12[1] = T2[1] - ynew;
    relativeT12[2] = T2[2] - znew ;

}

const double PI = 3.141592654; // PI value

// generate random number for the range (min, max)
double drand(double min, double max)
{
    return ((double)rand() / ((double)RAND_MAX + 1.0)) * (max - min) + min;
}

void generateUniformTranslation(vector3 &t, const double radius) {

    double theta = 2*PI*drand(0,1);
    double phi = acos(2*drand(0,1)-1.0);

    // now, compute the random radius
    double r = pow(drand(0,1),1./3.)*radius;

    // in a unit ball
    t[0] = r*cos(theta)*sin(phi);
    t[1] = r*sin(theta)*sin(phi);
    t[2] = r*cos(phi);
}

template<const int N>
struct dockingPose {
    quat q; // rigid quaternion rotation
    vector3 T; // rigid translation
    double flex[N]; // flexible amplitudes
};

double getLinearRMSD(vector3 *coords, double* q, vector3 T, int N, double *weights=NULL) {

    double Mass = 0 ;
    double A = q[3]*q[3]-(q[0]*q[0]+q[1]*q[1]+q[2]*q[2]);

    double r = 0.0;
    double x,y,z;
    vector3 rtmp;
    if (weights) {
        for (int i=0; i<N; ++i) {

            Mass+=weights[i];

            x=coords[i][0];
            y=coords[i][1];
            z=coords[i][2];


            double B = (x*q[0]+y*q[1]+z*q[2]);
            double cx = q[1]*z-q[2]*y;
            double cy = q[2]*x-q[0]*z;
            double cz = q[0]*y-q[1]*x;

            rtmp[0] = A*x+2*q[3]*cx+2*B*q[0];
            rtmp[1] = A*y+2*q[3]*cy+2*B*q[1];
            rtmp[2] = A*z+2*q[3]*cz+2*B*q[2];

            rtmp[0]+=T[0];
            rtmp[1]+=T[1];
            rtmp[2]+=T[2];

            r+= weights[i]* ((x-rtmp[0])*(x-rtmp[0])+(y-rtmp[1])*(y-rtmp[1])+(z-rtmp[2])*(z-rtmp[2]));
        }
        r/=Mass;
    }
    else {
        for (int i=0; i<N; ++i) {

            x=coords[i][0];
            y=coords[i][1];
            z=coords[i][2];

            double B = (x*q[0]+y*q[1]+z*q[2]);
            double cx = q[1]*z-q[2]*y;
            double cy = q[2]*x-q[0]*z;
            double cz = q[0]*y-q[1]*x;

            rtmp[0] = A*x+2*q[3]*cx+2*B*q[0];
            rtmp[1] = A*y+2*q[3]*cy+2*B*q[1];
            rtmp[2] = A*z+2*q[3]*cz+2*B*q[2];

            rtmp[0]+=T[0];
            rtmp[1]+=T[1];
            rtmp[2]+=T[2];

            r+= ((x-rtmp[0])*(x-rtmp[0])+(y-rtmp[1])*(y-rtmp[1])+(z-rtmp[2])*(z-rtmp[2]));
        }
        r/=N;
    }

    return r;
}

double getLinearRMSD(vector3 *coords1, vector3 *coords2, double* q, vector3 T, int N, double *weights=NULL) {
    
    double Mass = 0.0 ;
    double A = q[3]*q[3]-(q[0]*q[0]+q[1]*q[1]+q[2]*q[2]);
    
    double r = 0.0;
    double x,y,z;
    vector3 rtmp;
    if (weights) {
        for (int i=0; i<N; ++i) {
            
            Mass+=weights[i];
            
            x=coords2[i][0];
            y=coords2[i][1];
            z=coords2[i][2];
            
            
            double B = (x*q[0]+y*q[1]+z*q[2]);
            double cx = q[1]*z-q[2]*y;
            double cy = q[2]*x-q[0]*z;
            double cz = q[0]*y-q[1]*x;
            
            rtmp[0] = A*x+2*q[3]*cx+2*B*q[0];
            rtmp[1] = A*y+2*q[3]*cy+2*B*q[1];
            rtmp[2] = A*z+2*q[3]*cz+2*B*q[2];
            
            rtmp[0]+=T[0];
            rtmp[1]+=T[1];
            rtmp[2]+=T[2];
            
            x=coords1[i][0];
            y=coords1[i][1];
            z=coords1[i][2];

            r+= weights[i]* ((x-rtmp[0])*(x-rtmp[0])+(y-rtmp[1])*(y-rtmp[1])+(z-rtmp[2])*(z-rtmp[2]));
        }
        r/=Mass;
    }
    else {
        for (int i=0; i<N; ++i) {
            
            x=coords2[i][0];
            y=coords2[i][1];
            z=coords2[i][2];
            
            double B = (x*q[0]+y*q[1]+z*q[2]);
            double cx = q[1]*z-q[2]*y;
            double cy = q[2]*x-q[0]*z;
            double cz = q[0]*y-q[1]*x;
            
            rtmp[0] = A*x+2*q[3]*cx+2*B*q[0];
            rtmp[1] = A*y+2*q[3]*cy+2*B*q[1];
            rtmp[2] = A*z+2*q[3]*cz+2*B*q[2];
            
            rtmp[0]+=T[0];
            rtmp[1]+=T[1];
            rtmp[2]+=T[2];
            
            x=coords1[i][0];
            y=coords1[i][1];
            z=coords1[i][2];
            
            r+= ((x-rtmp[0])*(x-rtmp[0])+(y-rtmp[1])*(y-rtmp[1])+(z-rtmp[2])*(z-rtmp[2]));
        }
        r/=N;
    }

    return r;
}
double getLinearRMSD(vector3 *coords1,vector3 *coords2, double* m1, double *m2, vector3 T1, vector3 T2, int N, double *weights=NULL) {
    
    double Mass = 0 ;
    
    double r = 0;
    double x,y,z;
    vector3 r1,r2;
    if (weights) {
        for (int i=0; i<N; ++i) {
            
            Mass+=weights[i];
            
            x=coords1[i][0];
            y=coords1[i][1];
            z=coords1[i][2];
            
            r1[0] = m1[0]*x+m1[1]*y+m1[2]*z;
            r1[1] = m1[3]*x+m1[4]*y+m1[5]*z;
            r1[2] = m1[6]*x+m1[7]*y+m1[8]*z;
            
            r1[0]+=T1[0];
            r1[1]+=T1[1];
            r1[2]+=T1[2];
            
            x=coords2[i][0];
            y=coords2[i][1];
            z=coords2[i][2];
            
            r2[0] = m2[0]*x+m2[1]*y+m2[2]*z;
            r2[1] = m2[3]*x+m2[4]*y+m2[5]*z;
            r2[2] = m2[6]*x+m2[7]*y+m2[8]*z;
            
            r2[0]+=T2[0];
            r2[1]+=T2[1];
            r2[2]+=T2[2];
            
            r+= weights[i]* ((r1[0]-r2[0])*(r1[0]-r2[0])+(r1[1]-r2[1])*(r1[1]-r2[1])+(r1[2]-r2[2])*(r1[2]-r2[2]));
        }
        r/=Mass;
    }
    else {
        for (int i=0; i<N; ++i) {
            
            x=coords1[i][0];
            y=coords1[i][1];
            z=coords1[i][2];
            
            r1[0] = m1[0]*x+m1[1]*y+m1[2]*z;
            r1[1] = m1[3]*x+m1[4]*y+m1[5]*z;
            r1[2] = m1[6]*x+m1[7]*y+m1[8]*z;
            
            r1[0]+=T1[0];
            r1[1]+=T1[1];
            r1[2]+=T1[2];
            
            x=coords2[i][0];
            y=coords2[i][1];
            z=coords2[i][2];
            
            r2[0] = m2[0]*x+m2[1]*y+m2[2]*z;
            r2[1] = m2[3]*x+m2[4]*y+m2[5]*z;
            r2[2] = m2[6]*x+m2[7]*y+m2[8]*z;
            
            r2[0]+=T2[0];
            r2[1]+=T2[1];
            r2[2]+=T2[2];
            
            r+= ((r1[0]-r2[0])*(r1[0]-r2[0])+(r1[1]-r2[1])*(r1[1]-r2[1])+(r1[2]-r2[2])*(r1[2]-r2[2]));
        }
        r/=N;
    }
    return r;
}

double getLinearRMSD(vector3 *coords, double* q, vector3 T, int N,  vector3 **modes, const double *modes1, const double *modes2, int nModes, double *weights=NULL) {
    
    double Mass = 0 ;
    double A = q[3]*q[3]-(q[0]*q[0]+q[1]*q[1]+q[2]*q[2]);
    
    double r = 0;
    double x,y,z;
    double x1, y1, z1;
    vector3 rtmp;
    
    if (weights) {
        for (int i=0; i<N; ++i) {
            
            Mass+=weights[i];
            
            x1=coords[i][0];
            y1=coords[i][1];
            z1=coords[i][2];
            
            x=coords[i][0];
            y=coords[i][1];
            z=coords[i][2];
            
            for (int j = 0; j <nModes ; ++j){
                x1 += modes1[j]*modes[j][i][0];
                y1 += modes1[j]*modes[j][i][1];
                z1 += modes1[j]*modes[j][i][2];
                x += modes2[j]*modes[j][i][0];
                y += modes2[j]*modes[j][i][1];
                z += modes2[j]*modes[j][i][2];
            }
            
            
            double B = (x*q[0]+y*q[1]+z*q[2]);
            double cx = q[1]*z-q[2]*y;
            double cy = q[2]*x-q[0]*z;
            double cz = q[0]*y-q[1]*x;
            
            rtmp[0] = A*x+2*q[3]*cx+2*B*q[0];
            rtmp[1] = A*y+2*q[3]*cy+2*B*q[1];
            rtmp[2] = A*z+2*q[3]*cz+2*B*q[2];
            
            rtmp[0]+=T[0];
            rtmp[1]+=T[1];
            rtmp[2]+=T[2];
            
            r+= weights[i]* ((x1-rtmp[0])*(x1-rtmp[0])+(y1-rtmp[1])*(y1-rtmp[1])+(z1-rtmp[2])*(z1-rtmp[2]));
        }
        r/=Mass;
    }
    else {
        for (int i=0; i<N; ++i) {
            
            x1=coords[i][0];
            y1=coords[i][1];
            z1=coords[i][2];
            
            x=coords[i][0];
            y=coords[i][1];
            z=coords[i][2];
            
            for (int j = 0; j <nModes ; ++j){
                x1 += modes1[j]*modes[j][i][0];
                y1 += modes1[j]*modes[j][i][1];
                z1 += modes1[j]*modes[j][i][2];
                x += modes2[j]*modes[j][i][0];
                y += modes2[j]*modes[j][i][1];
                z += modes2[j]*modes[j][i][2];
            }
            
            double B = (x*q[0]+y*q[1]+z*q[2]);
            double cx = q[1]*z-q[2]*y;
            double cy = q[2]*x-q[0]*z;
            double cz = q[0]*y-q[1]*x;
            
            rtmp[0] = A*x+2*q[3]*cx+2*B*q[0];
            rtmp[1] = A*y+2*q[3]*cy+2*B*q[1];
            rtmp[2] = A*z+2*q[3]*cz+2*B*q[2];
            
            rtmp[0]+=T[0];
            rtmp[1]+=T[1];
            rtmp[2]+=T[2];
            
            r+= ((x1-rtmp[0])*(x1-rtmp[0])+(y1-rtmp[1])*(y1-rtmp[1])+(z1-rtmp[2])*(z1-rtmp[2]));
        }
        r/=N;
    }
    return r;
}



double getLinearRMSD(vector3 *coords, double* q1, double* q2, vector3 T1,   vector3 T2, int N,  vector3 **modes, const double *modes1, const double *modes2, int nModes, double *weights=NULL) {

    double Mass = 0 ;
    double A1 = q1[3]*q1[3]-(q1[0]*q1[0]+q1[1]*q1[1]+q1[2]*q1[2]);
    double A2 = q2[3]*q2[3]-(q2[0]*q2[0]+q2[1]*q2[1]+q2[2]*q2[2]);
    double r = 0;
    double x,y,z;
    double x1, y1, z1;
    vector3 rtmp1, rtmp2;

    if (weights) {
        for (int i=0; i<N; ++i) {

            Mass+=weights[i];

            x1=coords[i][0];
            y1=coords[i][1];
            z1=coords[i][2];

            x=coords[i][0];
            y=coords[i][1];
            z=coords[i][2];

            for (int j = 0; j <nModes ; ++j){
                x1 += modes1[j]*modes[j][i][0];
                y1 += modes1[j]*modes[j][i][1];
                z1 += modes1[j]*modes[j][i][2];
                x += modes2[j]*modes[j][i][0];
                y += modes2[j]*modes[j][i][1];
                z += modes2[j]*modes[j][i][2];
            }

            double B2 = (x*q2[0]+y*q2[1]+z*q2[2]);
            double cx = q2[1]*z-q2[2]*y;
            double cy = q2[2]*x-q2[0]*z;
            double cz = q2[0]*y-q2[1]*x;

            rtmp2[0] = A2*x+2*q2[3]*cx+2*B2*q2[0];
            rtmp2[1] = A2*y+2*q2[3]*cy+2*B2*q2[1];
            rtmp2[2] = A2*z+2*q2[3]*cz+2*B2*q2[2];

            rtmp2[0]+=T2[0];
            rtmp2[1]+=T2[1];
            rtmp2[2]+=T2[2];

            
            double B1 = (x1*q1[0]+y1*q1[1]+z1*q1[2]);
             cx = q1[1]*z1-q1[2]*y1;
             cy = q1[2]*x1-q1[0]*z1;
             cz = q1[0]*y1-q1[1]*x1;
            
            rtmp1[0] = A1*x1+2*q1[3]*cx+2*B1*q1[0];
            rtmp1[1] = A1*y1+2*q1[3]*cy+2*B1*q1[1];
            rtmp1[2] = A1*z1+2*q1[3]*cz+2*B1*q1[2];
            
            rtmp1[0]+=T1[0];
            rtmp1[1]+=T1[1];
            rtmp1[2]+=T1[2];

            
            r+= weights[i]* ((rtmp1[0]-rtmp2[0])*(rtmp1[0]-rtmp2[0])+(rtmp1[1]-rtmp2[1])*(rtmp1[1]-rtmp2[1])+(rtmp1[2]-rtmp2[2])*(rtmp1[2]-rtmp2[2]));
        }
        r/=Mass;
    }
    else {
        for (int i=0; i<N; ++i) {

            x1=coords[i][0];
            y1=coords[i][1];
            z1=coords[i][2];

            x=coords[i][0];
            y=coords[i][1];
            z=coords[i][2];

            for (int j = 0; j <nModes ; ++j){
                x1 += modes1[j]*modes[j][i][0];
                y1 += modes1[j]*modes[j][i][1];
                z1 += modes1[j]*modes[j][i][2];
                x += modes2[j]*modes[j][i][0];
                y += modes2[j]*modes[j][i][1];
                z += modes2[j]*modes[j][i][2];
            }

            double B2 = (x*q2[0]+y*q2[1]+z*q2[2]);
            double cx = q2[1]*z-q2[2]*y;
            double cy = q2[2]*x-q2[0]*z;
            double cz = q2[0]*y-q2[1]*x;
            
            rtmp2[0] = A2*x+2*q2[3]*cx+2*B2*q2[0];
            rtmp2[1] = A2*y+2*q2[3]*cy+2*B2*q2[1];
            rtmp2[2] = A2*z+2*q2[3]*cz+2*B2*q2[2];
            
            rtmp2[0]+=T2[0];
            rtmp2[1]+=T2[1];
            rtmp2[2]+=T2[2];
            
            
            double B1 = (x1*q2[0]+y1*q2[1]+z1*q2[2]);
            cx = q1[1]*z-q1[2]*y1;
            cy = q1[2]*x-q1[0]*z1;
            cz = q1[0]*y-q1[1]*x1;
            
            rtmp1[0] = A1*x1+2*q1[3]*cx+2*B1*q1[0];
            rtmp1[1] = A1*y1+2*q1[3]*cy+2*B1*q1[1];
            rtmp1[2] = A1*z1+2*q1[3]*cz+2*B1*q1[2];
            
            rtmp1[0]+=T1[0];
            rtmp1[1]+=T1[1];
            rtmp1[2]+=T1[2];
            
            
            r+= ((rtmp1[0]-rtmp2[0])*(rtmp1[0]-rtmp2[0])+(rtmp1[1]-rtmp2[1])*(rtmp1[1]-rtmp2[1])+(rtmp1[2]-rtmp2[2])*(rtmp1[2]-rtmp2[2]));

        }
        r/=N;
    }
    return r;
}

double getLinearRMSD(vector3 *coords, double* q, vector3 T, int N,  vector3 **modes, const double *modes1, int nModes, double *weights=NULL) {

    double Mass = 0 ;
    double A = q[3]*q[3]-(q[0]*q[0]+q[1]*q[1]+q[2]*q[2]);


    double r = 0;
    double x,y,z;
    double x1, y1, z1;
    vector3 rtmp;

    if (weights) {
        for (int i=0; i<N; ++i) {

            Mass+=weights[i];

            x1=coords[i][0];
            y1=coords[i][1];
            z1=coords[i][2];

            x=coords[i][0];
            y=coords[i][1];
            z=coords[i][2];

            for (int j = 0; j <nModes ; ++j){
                x1 += modes1[j]*modes[j][i][0];
                y1 += modes1[j]*modes[j][i][1];
                z1 += modes1[j]*modes[j][i][2];
            }

            double B = (x*q[0]+y*q[1]+z*q[2]);
            double cx = q[1]*z-q[2]*y;
            double cy = q[2]*x-q[0]*z;
            double cz = q[0]*y-q[1]*x;

            rtmp[0] = A*x+2*q[3]*cx+2*B*q[0];
            rtmp[1] = A*y+2*q[3]*cy+2*B*q[1];
            rtmp[2] = A*z+2*q[3]*cz+2*B*q[2];

            rtmp[0]+=T[0];
            rtmp[1]+=T[1];
            rtmp[2]+=T[2];

            r+= weights[i]* ((x1-rtmp[0])*(x1-rtmp[0])+(y1-rtmp[1])*(y1-rtmp[1])+(z1-rtmp[2])*(z1-rtmp[2]));
        }
        r/=Mass;
    }
    else {
        for (int i=0; i<N; ++i) {

            x1=coords[i][0];
            y1=coords[i][1];
            z1=coords[i][2];

            x=coords[i][0];
            y=coords[i][1];
            z=coords[i][2];

            for (int j = 0; j <nModes ; ++j){
                x1 += modes1[j]*modes[j][i][0];
                y1 += modes1[j]*modes[j][i][1];
                z1 += modes1[j]*modes[j][i][2];
            }

            double B = (x*q[0]+y*q[1]+z*q[2]);
            double cx = q[1]*z-q[2]*y;
            double cy = q[2]*x-q[0]*z;
            double cz = q[0]*y-q[1]*x;

            rtmp[0] = A*x+2*q[3]*cx+2*B*q[0];
            rtmp[1] = A*y+2*q[3]*cy+2*B*q[1];
            rtmp[2] = A*z+2*q[3]*cz+2*B*q[2];

            rtmp[0]+=T[0];
            rtmp[1]+=T[1];
            rtmp[2]+=T[2];

            r+= ((x1-rtmp[0])*(x1-rtmp[0])+(y1-rtmp[1])*(y1-rtmp[1])+(z1-rtmp[2])*(z1-rtmp[2]));
        }
        r/=N;
    }
    return r;
}

void readCoordinates(vector3 *&coords, double *&weights, int &size, double &mass) {

    //test coordinates
    //initialization with pdbLines
    std::vector<cPDBAtom> atoms;
    atoms.resize(20);

    //first, we create a 20-atoms template
    atoms[0].initialize("ATOM      4  N   GLU A   1       1.941  -2.006   3.760  1.00 25.22           N");
    atoms[1].initialize("ATOM      5  CA  GLU A   1       2.783  -2.083   2.553  1.00 29.88           C");
    atoms[2].initialize("ATOM      6  C   GLU A   1       2.110  -2.700   2.235  1.00 26.79           C");
    atoms[3].initialize("ATOM      7  O   GLU A   1       2.801  -2.237   1.337  1.00 22.16           O");
    atoms[4].initialize("ATOM      8  CB  GLU A   1       3.276  -2.495   2.288  1.00 25.90           C");
    atoms[5].initialize("ATOM      9  CG  GLU A   1       3.603  -0.188   3.574  1.00 48.92           C");
    atoms[6].initialize("ATOM     10  CD  GLU A   1       2.380  -0.852   4.025  1.00 29.36           C");
    atoms[7].initialize("ATOM     11  OE1 GLU A   1       1.978   1.935   3.576  1.00 28.67           O");
    atoms[8].initialize("ATOM     12  OE2 GLU A   1       1.821  -0.061   4.894  1.00 37.04           O");
    atoms[9].initialize("ATOM     13  N   TRP A   2       0.800  -2.015   2.074  1.00 20.72           N");
    atoms[10].initialize("ATOM     14  CA  TRP A   2       0.045  -3.649  -1.874  1.00 16.62           C");
    atoms[11].initialize("ATOM     15  C   TRP A   2      -0.264  -4.184  -1.771  1.00 18.21           C");
    atoms[12].initialize("ATOM     16  O   TRP A   2      -0.183  -5.585  -2.710  1.00 31.39           O");
    atoms[13].initialize("ATOM     17  CB  TRP A   2      -1.170  -2.520  -1.699  1.00 21.45           C");
    atoms[14].initialize("ATOM     18  CG  TRP A   2      -0.762  -1.955  -1.503  1.00 23.04           C");
    atoms[15].initialize("ATOM     19  CD1 TRP A   2      -0.834  -0.953   0.417  1.00 30.95           C");
    atoms[16].initialize("ATOM     20  CD2 TRP A   2      -0.289  -0.563  -2.315  1.00 18.69           C");
    atoms[17].initialize("ATOM     21  NE1 TRP A   2      -0.413  -2.140  -2.903  1.00 24.73           N");
    atoms[18].initialize("ATOM     22  CE2 TRP A   2      -0.203  -1.945  -3.564  1.00 27.52           C");
    atoms[19].initialize("ATOM     23  CE3 TRP A   2      -0.126  -0.093  -3.056  1.00 20.75           C");

    //then, we will randomly replicate its size/20 times
    int templateSize = atoms.size();
    double amplitude = pow(10*size, 1./3); // about water density
    for (int i=0; i<size/templateSize-1; ++i) {
        for (int j =0; j<templateSize; ++j ) {
            cPDBAtom newAtom = atoms[j];
            vector3 randomDisplacement;
            generateUniformTranslation(randomDisplacement, amplitude);
            newAtom.coords[0] += randomDisplacement[0];
            newAtom.coords[1] += randomDisplacement[1];
            newAtom.coords[2] += randomDisplacement[2];
            atoms.push_back(newAtom);
        }
    }

    std::cout<<atoms.size()<<" atoms created with displacements in a ball of radius "<<amplitude<<" A "<<std::endl;

    size = atoms.size();
    coords = new vector3[atoms.size()];
    weights = new double[atoms.size()];

    for (unsigned int i=0; i<atoms.size(); i++) {
        coords[i][0] = atoms[i].coords[0];
        coords[i][1] = atoms[i].coords[1];
        coords[i][2] = atoms[i].coords[2];
        weights[i] = atoms[i].weight;
        mass +=atoms[i].weight;
    }
    std::cout<<" mass "<< mass  <<" a.u."<<std::endl;
}

void readWeightOrthogonalModes (vector3 **modes, int nModes, int size, double *weights = NULL) {

    double modenorm;
    double projection;


    for (int j = 0; j<nModes ; ++j){
        modes[j] = new vector3[size];

        modenorm=0.0;
        for (int i=0; i<size; i++) {

            generateUniformTranslation( modes[j][i], 1);

        }

        //Gram–Schmidt projections to the previous modes
        for (int k =0; k< j; ++k) {
            projection = 0.0;
            for (int i=0; i<size; i++) {
                projection +=  weights[i]*(modes[j][i][0]* modes[k][i][0] +modes[j][i][1]* modes[k][i][1]+ modes[j][i][2]* modes[k][i][2]);
            }
            for (int i=0; i<size; i++) {
                modes[j][i][0] -= projection*modes[k][i][0];
                modes[j][i][1] -= projection*modes[k][i][1];
                modes[j][i][2] -= projection*modes[k][i][2];
            }
        }

        for (unsigned int i=0; i<size; i++) {
            modenorm += weights[i]*(modes[j][i][0]* modes[j][i][0] +modes[j][i][1]* modes[j][i][1]+ modes[j][i][2]* modes[j][i][2]);
        }
        modenorm = sqrt(modenorm);

        for (int i=0; i<size; i++) {
            modes[j][i][0]/=modenorm;
            modes[j][i][1]/=modenorm;
            modes[j][i][2]/=modenorm;

        }
    }

    std::cout<<nModes<<" weight-orthonormal deformations created"<<std::endl;

}

void weightOrthogonalModesTest (vector3 **modes, int nModes, int size, double *weights) {

    // orthogonality test

    for (int j = 0; j<nModes;  ++j){

        double modenorm = 0.0;

        for (unsigned int i=0; i<size; i++) {
            modenorm += weights[i]*(modes[j][i][0]* modes[j][i][0] +modes[j][i][1]* modes[j][i][1]+ modes[j][i][2]* modes[j][i][2]);
        }
        modenorm = sqrt(modenorm);

        std::cout<< "modes "<<j << " ; norm = " << modenorm<<std::endl;
    }

    for (int j = 0; j<nModes;  ++j){
        for (int k = j+1; k<nModes;  ++k){
            double crossmode =0.0;

            for (unsigned int i=0; i<size; i++) {
                crossmode +=  weights[i]*(modes[j][i][0]* modes[k][i][0] +modes[j][i][1]* modes[k][i][1]+ modes[j][i][2]* modes[k][i][2]);
            }
            std::cout<< "cross-modes  "<<j << "  "<< k << "  = " << crossmode<<std::endl;
            
        }
    }

}

void readModes (vector3 **modes, int nModes, int size, double *weights = NULL) {

    double modenorm;
    double crossmode;


    for (int j = 0; j<nModes ; ++j){
        modes[j] = new vector3[size];

        modenorm=0.0;
        for (int i=0; i<size; i++) {
            generateUniformTranslation( modes[j][i], 1);
        }

        for (unsigned int i=0; i<size; i++) {
            modenorm += weights[i]*(modes[j][i][0]* modes[j][i][0] +modes[j][i][1]* modes[j][i][1]+ modes[j][i][2]* modes[j][i][2]);
        }

        modenorm = sqrt(modenorm);

        for (int i=0; i<size; i++) {
            modes[j][i][0]/=modenorm;
            modes[j][i][1]/=modenorm;
            modes[j][i][2]/=modenorm;

        }
    }
    std::cout<<nModes<<" modes created for each atoms of each proteins"<<std::endl;

    // orthogonality test

    for (int j = 0; j<nModes;  ++j){

        modenorm = 0.0;

        for (unsigned int i=0; i<size; i++) {
            modenorm += weights[i]*(modes[j][i][0]* modes[j][i][0] +modes[j][i][1]* modes[j][i][1]+ modes[j][i][2]* modes[j][i][2]);
        }
        modenorm = sqrt(modenorm);

        std::cout<< "modes "<<j << " ; norm = " << modenorm<<std::endl;
    }

    for (int j = 0; j<nModes;  ++j){
        for (int k = j+1; k<nModes;  ++k){
            crossmode =0.0;

            for (unsigned int i=0; i<size; i++) {
                crossmode +=  weights[i]*(modes[j][i][0]* modes[k][i][0] +modes[j][i][1]* modes[k][i][1]+ modes[j][i][2]* modes[k][i][2]);
            }
            std::cout<< "cross-modes  "<<j << "  "<< k << "  = " << crossmode<<std::endl;

        }
    }

}


void generateRigidTransformation(double q[], vector3 &T, double ampl = 1.0) {

    double qLength;

    generateUniformTranslation(T, ampl);

    vector3 n;
    generateUniformTranslation(n,1);

    q[0]= n[0];
    q[1]= n[1];
    q[2]= n[2];
    q[3]= drand(-1,1);

    //length of net quaternion
    qLength = sqrt(q[0]*q[0]+q[1]*q[1]+q[2]*q[2]+q[3]*q[3]);
    //normalize it
    q[0]/=qLength;
    q[1]/=qLength;
    q[2]/=qLength;
    q[3]/=qLength;


}

void generateModeAmplitudes (double mode1amp[] , double mode2amp[], int nmodes) {

    for (int j=0; j<nmodes; ++j) {

        mode1amp[j] = drand(0, 100);
        mode2amp[j] = drand(0, 100);
    }
}

void generateModeAmplitudes (double mode1amp[] , int nmodes) {

    for (int j=0; j<nmodes; ++j) {

        mode1amp[j]= drand(0, 100);
    }
}

template<const int nmodes>
void readDockingPoses(dockingPose<nmodes> *poses, int nDockingPoses) {

    for (int i=0; i<nDockingPoses; ++i) {
        generateRigidTransformation(poses[i].q, poses[i].T, 100.);
        generateModeAmplitudes (poses[i].flex, nmodes);
    }
}


template<const int nmodes>
void generatePertubatedCoords(dockingPose<nmodes> *poses, int nDockingPoses, vector3 * coords, vector3 ** pertubatedCoords, int size, vector3 **modes, int nbmodes) {
    
    
    for ( int p = 0; p < nDockingPoses; ++p){ // loop on docking pose
        pertubatedCoords[p] = new vector3[size];

        for (int n =0; n<size; ++n) {// for on atoms
            
            pertubatedCoords[p][n][0] = coords[n][0] ;
            pertubatedCoords[p][n][1] = coords[n][1] ;
            pertubatedCoords[p][n][2] = coords[n][2] ;
            
            for (int i=0; i<nbmodes; ++i) {// for on modes
                pertubatedCoords[p][n][0] +=   poses[p].flex[i]*modes[i][n][0];
                pertubatedCoords[p][n][1] +=   poses[p].flex[i]*modes[i][n][1];
                pertubatedCoords[p][n][2] +=   poses[p].flex[i]*modes[i][n][2];
   
               
            } // for on modes
            
        } // for on atoms
    } // for on docking pose
    
}

//test only
template<const int nmodes>
void generatetransformedCoords(dockingPose<nmodes> poses , vector3 * coords, vector3 *&transformedCoords, int size) {

    double x,y,z;

    transformedCoords = new vector3[size];

    double q[4];
    q[0]= poses.q[0];  q[1]= poses.q[1];  q[2]= poses.q[2];  q[3]= poses.q[3];

    double T[3];
    T[0]= poses.T[0];     T[1]= poses.T[1];     T[2]= poses.T[2];     T[3]= poses.T[3];

    double A = q[3]*q[3]-(q[0]*q[0]+q[1]*q[1]+q[2]*q[2]);

    for (int n =0; n<size; ++n) {// for on atoms
        x=coords[n][0];
        y=coords[n][1];
        z=coords[n][2];

        double B = (x*q[0]+y*q[1]+z*q[2]);
        double cx = q[1]*z-q[2]*y;
        double cy = q[2]*x-q[0]*z;
        double cz = q[0]*y-q[1]*x;

        transformedCoords[n][0] = A*x+2*q[3]*cx+2*B*q[0];
        transformedCoords[n][1]  = A*y+2*q[3]*cy+2*B*q[1];
        transformedCoords[n][2] = A*z+2*q[3]*cz+2*B*q[2];

        transformedCoords[n][0]+=T[0];
        transformedCoords[n][1] +=T[1];
        transformedCoords[n][2]+=T[2];



        } // for on atoms
}




template<const int nmodes>
double fastRmsdTest (RMSD<vector3> &rmsd, dockingPose<nmodes> &pose1, dockingPose<nmodes> &pose2) {

    
    double m1[9], m2[9];
    computeMatrix (m1, pose1.q);
    computeMatrix (m2, pose2.q);

    return  rmsd.getMSDTwoTransformsOrtho( m1, m2,pose1.T,pose2.T, pose1.flex, pose2.flex);
 
    }


// computes pertubated coords +rms
template<const int nmodes>
double slowRmsdTest (dockingPose<nmodes> &pose1, dockingPose<nmodes> &pose2, vector3 *coords, int size, vector3 **modes, int nbmodes, double *weights = NULL) {

    
    return getLinearRMSD(coords, pose1.q, pose2.q, pose1.T,pose2.T,size, modes, pose1.flex, pose2.flex, nbmodes, weights);
}

// computes rms from pertubated coords
template<const int nmodes>
double slowRmsdTest (dockingPose<nmodes> &pose1, dockingPose<nmodes> &pose2, vector3 *coords1, vector3 *coords2 , int size, double *weights = NULL) {
    

    double m1[9] , m2[9];
    computeMatrix (m1, pose1.q);
    computeMatrix (m2, pose2.q);

    return getLinearRMSD(coords1, coords2, m1, m2,pose1.T,pose2.T,size, weights);
}

void clear(vector3 *coords, double *weights, vector3 **modes, int nModes) {
    delete [] coords;
    delete [] weights;

    for (int i=0; i< nModes; ++i)
        delete [] modes[i];

    delete [] modes;
}

int main() {

    //srand( time(NULL) );

    std::cout<<"Clustering example\n";

    /*--------------------------------------------------------------------------------------*/
    /*             Initialization of the example                                            */
    /*--------------------------------------------------------------------------------------*/
    // timer
    timer t;

    t.start();
    //coordinates and weights of the molecule
    vector3 *coords;
    double *weights;
    
    std::vector<int>  size;
    size.push_back(1000);
    size.push_back(5000);
    size.push_back(10000);
    size.push_back(20000);
    size.push_back(50000);
    size.push_back(100000);

    
    


    // mass of the molecule
    double mass =0.0;

    // desired number of particles, randomly replicated from a 20-atoms template
    int Natoms = size.back();

    // init collective motions
    const int nmodes = 50;
    vector3 **modes = new vector3*[nmodes];

    readCoordinates(coords, weights, Natoms, mass); // randomly replicated from a 20-atoms template
    readWeightOrthogonalModes(modes, nmodes, Natoms, weights); // random weight-orthonormal modes
//    weightOrthogonalModesTest(modes, nmodes, size, weights); // test the weight-orthonormal modes

    const int nDockingPoses =1000; // number of docking poses to cluster

    dockingPose<nmodes> poses[nDockingPoses]; // allocation of nDockingPoses
    readDockingPoses<nmodes>(poses, nDockingPoses); // random poses generated
    t.stop("Input Data generation");

    /*--------------------------------------------------------------------------------------*/
    /*                          Actual clustering                                           */
    /*--------------------------------------------------------------------------------------*/

    double thresholdRMSD = 120;

    std::vector<std::vector<int> > solutions; //output of the clustering method,
                                              //stores only the indices of the docking solutions

    std::vector<bool> clusteredFlag(nDockingPoses, false); // array that indicates if the pose has
                                                           // been already clustered
    
    
 


    for (int n= 0; n<size.size(); n++){ //loop on different nb of atoms
        std::ostringstream convert;
        convert << size[n];
        //creates files for output
        std::string Result = convert.str(); //
        std::string file =   "Results/Atoms_"+Result+".txt";
        std::ofstream myfile (file);
        
        for(int nbmodes =5 ; nbmodes<=50; nbmodes +=5){
            
            std::ostringstream convertModes;
            convertModes << nbmodes;
            //creates files for output
            std::string Result2 = convertModes.str(); //
            std::string file2 =   "Results/Modes_"+Result2+".txt";
            std::ofstream myfile2;
            
            myfile2.open (file2, std::ofstream::ate | std::ofstream::app);

            
            vector3 **perturbatedCoords = new vector3*[nDockingPoses]; //pertubated coords calculated once before clustering tests

            std::cout<<"\nTest with "<< size[n] <<" atoms and "<< nbmodes <<" modes \n";
            
            /*--------------------------------------------------------------------------------------*/
            /*                            Slow clustering                                           */
            /*--------------------------------------------------------------------------------------*/
            t.start();
            double thresholdRMSD2 = thresholdRMSD*thresholdRMSD;
            for (int seed = 0; seed<nDockingPoses; ++ seed) {
                if (clusteredFlag[seed] == true) continue; // skip already clustered solutions
                solutions.push_back(std::vector<int>(seed));
                clusteredFlag[seed] = true;
                for (int target = seed+1; target < nDockingPoses; ++ target) {
                    if (clusteredFlag[target] == true) continue; // skip already clustered solutions
                    if (slowRmsdTest(poses[seed], poses[target], coords, size[n], modes,nbmodes, weights) < thresholdRMSD2) {
                        solutions.back().push_back(target);
                        clusteredFlag[target] = true;
                    }
                }
            }
            double slowTime = t.stop("Slow Clustering");
            std::cout<<"Created "<<solutions.size()<<" cluster(s) out of "<<nDockingPoses<<" poses."<<std::endl;
            
            solutions.clear();
            clusteredFlag.clear();
            clusteredFlag.resize(nDockingPoses, false);
            
            /*--------------------------------------------------------------------------------------*/
            /*                            Slow clustering using pertubated coords                   */
            /*--------------------------------------------------------------------------------------*/
            
            
            // coordinates generated outside the loop
            generatePertubatedCoords<nmodes>(poses,nDockingPoses, coords, perturbatedCoords, size[n], modes, nbmodes);
    
            t.start();
            
            for (int seed = 0; seed<nDockingPoses; ++ seed) {
                if (clusteredFlag[seed] == true) continue; // skip already clustered solutions
                solutions.push_back(std::vector<int>(seed));
                clusteredFlag[seed] = true;
                for (int target = seed+1; target < nDockingPoses; ++ target) {
                    if (clusteredFlag[target] == true) continue; // skip already clustered solutions
                    if (slowRmsdTest(poses[seed], poses[target], perturbatedCoords[seed],perturbatedCoords[target], size[n], weights) < thresholdRMSD2) {
                        solutions.back().push_back(target);
                        clusteredFlag[target] = true;
                    }
                }
            }
            double slowTime2 = t.stop("Slow Clustering with peturbated coordinates ");
            std::cout<<"Created "<<solutions.size()<<" cluster(s) out of "<<nDockingPoses<<" poses."<<std::endl;
            solutions.clear();
            clusteredFlag.clear();
            clusteredFlag.resize(nDockingPoses, false);

            for (int i=0; i< nDockingPoses; ++i)
                delete [] perturbatedCoords[i];
            
            delete [] perturbatedCoords;

            
            /*--------------------------------------------------------------------------------------*/
            /*                          Rapid RMSD init                                             */
            /*--------------------------------------------------------------------------------------*/
            t.start();
            RMSD<vector3> rmsd;
            rmsd.initModes(coords, size[n], modes, nbmodes, weights); 	// flex RMSD init
            double initTime = t.stop("Rapid RMSD initialization");
            /*--------------------------------------------------------------------------------------*/
            /*                            Fast clustering                                           */
            /*--------------------------------------------------------------------------------------*/
            t.start();
          //  thresholdRMSD2 = thresholdRMSD*thresholdRMSD;
            for (int seed = 0; seed<nDockingPoses; ++ seed) {
                if (clusteredFlag[seed] == true) continue; // skip already clustered solutions
                solutions.push_back(std::vector<int>(seed));
                clusteredFlag[seed] = true;
                for (int target = seed+1; target < nDockingPoses; ++ target) {
                    if (clusteredFlag[target] == true) continue; // skip already clustered solutions
                    if (fastRmsdTest(rmsd, poses[seed], poses[target]) < thresholdRMSD2) {
                        solutions.back().push_back(target);
                        clusteredFlag[target] = true;
                    }
                }
            }
            double fastTime = t.stop("Fast Clustering");
            std::cout<<"Created "<<solutions.size()<<" cluster(s) out of "<<nDockingPoses<<" poses."<<std::endl;

            solutions.clear();
            clusteredFlag.clear();
            clusteredFlag.resize(nDockingPoses, false);
            
            std::cout<<"Asymptotic speedup for "<<size[n]<<" particles and "<<nbmodes<<" deformations is "<<slowTime/fastTime<<std::endl;
            std::cout<<"Asymptotic speedup using pertubated coordinates "<<size[n]<<" particles and "<<nbmodes<<" deformations is "<<slowTime2/fastTime<<std::endl;
        
            myfile << nbmodes << "       "<<fastTime<< "        "<< slowTime/fastTime << "      "<< slowTime2/fastTime <<"\n";
            myfile2<<size[n]<<"       "<<fastTime<< "        "<< slowTime/fastTime << "      "<< slowTime2/fastTime <<"\n";
            //stores the results
            myfile2.close();

        } // ef different nb of modes
        
        myfile.close();


        
    }//ef different nb of atoms
    

    
    

    clear(coords, weights, modes, nmodes);
}
