//Copyright INRIA/CNRS
//
//Neveu, Emilie; Popov, Petr; Hoffmann, Alexandre; Grudinin, Sergei. 2017
//sergei.grudinin@inria.fr
//
//This software is a computer program whose purpose is to compute
//the root mean sum of squared deviations (RMSDs) corresponding 
//to a rigid-body motion in constatnt time.
//
//This software is governed by the CeCILL-B  license (fully BSD-compatible)
//under French law and abiding by the rules of distribution of free software.
//You can  use, modify and/ or redistribute the software under the terms
//of the CeCILL-B license as circulated by CEA, CNRS and INRIA at the
//following URL "http://www.cecill.info".
//
//As a counterpart to the access to the source code and  rights to copy,
//modify and redistribute granted by the license, users are provided only
//with a limited warranty  and the software's author,  the holder of the
//economic rights,  and the successive licensors  have only  limited
//liability.
//
//In this respect, the user's attention is drawn to the risks associated
//with loading,  using,  modifying and/or developing or reproducing the
//software by the user in light of its specific status of free software,
//that may mean  that it is complicated to manipulate,  and  that  also
//therefore means  that it is reserved for developers  and  experienced
//professionals having in-depth computer knowledge. Users are therefore
//encouraged to load and test the software's suitability as regards their
//requirements in conditions enabling the security of their systems and/or
//data to be ensured and,  more generally, to use and operate it in the
//same conditions as regards security.
//
//The fact that you are presently reading this means that you have had
//knowledge of the CeCILL-B license and that you accept its terms.

/*
 *============================================
 *==== example of getMSDMatrix() function ====
 *============================================
 */

#include "RapidRMSD.hpp" // main library
#include <stdlib.h>
#include <vector>
#include "cPDBAtom.hpp" // for the PDB input
#include <math.h> // for sqrt
#include <time.h> // for srand

typedef double vector3[3]; 

void computeMatrix (double* m, double *q) {

	if (m && q) {
		double a = q[3];
		double b = q[0];
		double c = q[1];
		double d = q[2];

		m[0] = a*a+b*b-c*c-d*d; m[4] = a*a-b*b+c*c-d*d; m[8] = a*a-b*b-c*c+d*d;
		m[1] = 2*b*c-2*a*d; m[3] = 2*b*c+2*a*d;
		m[2] = 2*b*d+2*a*c; m[6] = 2*b*d-2*a*c;
		m[5] = 2*c*d-2*a*b; m[7] = 2*c*d+2*a*b;

		return;
	}
	else {
		std::cerr << "Error"<<std::endl;
		exit(0);
	}
}

double getLinearRMSD(vector3 *coords, double* m, vector3 T, int N, double *weights=NULL) {

	double Mass = 0 ;
	
	double msd = 0;
	double x,y,z;
	vector3 r;
	if (weights) {
		for (int i=0; i<N; ++i) {

			Mass+=weights[i];

			x=coords[i][0];
			y=coords[i][1];
			z=coords[i][2];

			r[0] = m[0]*x+m[1]*y+m[2]*z;
			r[1] = m[3]*x+m[4]*y+m[5]*z;
			r[2] = m[6]*x+m[7]*y+m[8]*z;

			r[0]+=T[0];
			r[1]+=T[1];
			r[2]+=T[2];

			msd+= weights[i]* ((x-r[0])*(x-r[0])+(y-r[1])*(y-r[1])+(z-r[2])*(z-r[2]));			
		}
		msd/=Mass;
	}
	else {
		for (int i=0; i<N; ++i) {

			x=coords[i][0];
			y=coords[i][1];
			z=coords[i][2];

			r[0] = m[0]*x+m[1]*y+m[2]*z;
			r[1] = m[3]*x+m[4]*y+m[5]*z;
			r[2] = m[6]*x+m[7]*y+m[8]*z;

			r[0]+=T[0];
			r[1]+=T[1];
			r[2]+=T[2];

			msd+= ((x-r[0])*(x-r[0])+(y-r[1])*(y-r[1])+(z-r[2])*(z-r[2]));			
		}
		msd/=N;
	}
	return msd;
}


int main() {

	srand ( time(NULL) );

	std::cout<<"example of getMSDMatrix() function\n";

	int N = 20;
	vector3 *coords = new vector3[N];
		
	//test coordinates
	//initialization with pdbLines
	std::vector<cPDBAtom> atoms;
	atoms.resize(20);

	atoms[0].initialize("ATOM      4  N   GLU A   1       1.941  -2.006   3.760  1.00 25.22           N");
	atoms[1].initialize("ATOM      5  CA  GLU A   1       2.783  -2.083   2.553  1.00 29.88           C");
	atoms[2].initialize("ATOM      6  C   GLU A   1       2.110  -2.700   2.235  1.00 26.79           C");
	atoms[3].initialize("ATOM      7  O   GLU A   1       2.801  -2.237   1.337  1.00 22.16           O");
	atoms[4].initialize("ATOM      8  CB  GLU A   1       3.276  -2.495   2.288  1.00 25.90           C");
	atoms[5].initialize("ATOM      9  CG  GLU A   1       3.603  -0.188   3.574  1.00 48.92           C");
	atoms[6].initialize("ATOM     10  CD  GLU A   1       2.380  -0.852   4.025  1.00 29.36           C");
	atoms[7].initialize("ATOM     11  OE1 GLU A   1       1.978   1.935   3.576  1.00 28.67           O");
	atoms[8].initialize("ATOM     12  OE2 GLU A   1       1.821  -0.061   4.894  1.00 37.04           O");
	atoms[9].initialize("ATOM     13  N   TRP A   2       0.800  -2.015   2.074  1.00 20.72           N");
	atoms[10].initialize("ATOM     14  CA  TRP A   2       0.045  -3.649  -1.874  1.00 16.62           C");
	atoms[11].initialize("ATOM     15  C   TRP A   2      -0.264  -4.184  -1.771  1.00 18.21           C");
	atoms[12].initialize("ATOM     16  O   TRP A   2      -0.183  -5.585  -2.710  1.00 31.39           O");
	atoms[13].initialize("ATOM     17  CB  TRP A   2      -1.170  -2.520  -1.699  1.00 21.45           C");
	atoms[14].initialize("ATOM     18  CG  TRP A   2      -0.762  -1.955  -1.503  1.00 23.04           C");
	atoms[15].initialize("ATOM     19  CD1 TRP A   2      -0.834  -0.953   0.417  1.00 30.95           C");
	atoms[16].initialize("ATOM     20  CD2 TRP A   2      -0.289  -0.563  -2.315  1.00 18.69           C");
	atoms[17].initialize("ATOM     21  NE1 TRP A   2      -0.413  -2.140  -2.903  1.00 24.73           N");
	atoms[18].initialize("ATOM     22  CE2 TRP A   2      -0.203  -1.945  -3.564  1.00 27.52           C");
	atoms[19].initialize("ATOM     23  CE3 TRP A   2      -0.126  -0.093  -3.056  1.00 20.75           C");
	
	std::cout<<atoms.size()<<" atoms created"<<std::endl;
	
	
	// compute coordinates, unit weights
	
	for (unsigned int i=0; i<atoms.size(); i++) {
		coords[i][0] = atoms[i].coords[0];
		coords[i][1] = atoms[i].coords[1];
		coords[i][2] = atoms[i].coords[2];
	}

	RMSD<vector3> rmsd;
	rmsd.init(coords, N); 	// RMSD init

	double qLength;
	double q[4];
	double m[9];
	vector3 T;
	
	for (int i=0; i<5; ++i) {
		T[0] = double(rand())/RAND_MAX-0.5;
		T[1] = double(rand())/RAND_MAX-0.5;
		T[2] = double(rand())/RAND_MAX-0.5;
		q[0]= double(rand())/RAND_MAX;
		q[1]= double(rand())/RAND_MAX;
		q[2]= double(rand())/RAND_MAX;
		q[3]= double(rand())/RAND_MAX;
		q[0]/=100; q[1]/=100; q[2]/=100; q[3]/=100;
		//length of net quaternion
		qLength = sqrt(q[0]*q[0]+q[1]*q[1]+q[2]*q[2]+q[3]*q[3]);
		//normalize it
		q[0]/=qLength;
		q[1]/=qLength;
		q[2]/=qLength;
		q[3]/=qLength;
		computeMatrix (m, q);

		std::cout<<"fast rmsd: "<<sqrt(rmsd.getMSDMatrix(m, T)); // RMSD function call
		std::cout<<" linear RMSD: "<<sqrt(getLinearRMSD(coords, m, T, N))<<std::endl; // comparison with a direct evaluation
	}
		
	delete [] coords;
}
