//Copyright INRIA/CNRS
//
//Neveu, Emilie; Popov, Petr; Hoffmann, Alexandre; Grudinin, Sergei. 2017
//sergei.grudinin@inria.fr
//
//This software is a computer program whose purpose is to compute
//the root mean sum of squared deviations (RMSDs) corresponding 
//to a rigid-body motion in constatnt time.
//
//This software is governed by the CeCILL-B  license (fully BSD-compatible)
//under French law and abiding by the rules of distribution of free software.
//You can  use, modify and/ or redistribute the software under the terms
//of the CeCILL-B license as circulated by CEA, CNRS and INRIA at the
//following URL "http://www.cecill.info".
//
//As a counterpart to the access to the source code and  rights to copy,
//modify and redistribute granted by the license, users are provided only
//with a limited warranty  and the software's author,  the holder of the
//economic rights,  and the successive licensors  have only  limited
//liability.
//
//In this respect, the user's attention is drawn to the risks associated
//with loading,  using,  modifying and/or developing or reproducing the
//software by the user in light of its specific status of free software,
//that may mean  that it is complicated to manipulate,  and  that  also
//therefore means  that it is reserved for developers  and  experienced
//professionals having in-depth computer knowledge. Users are therefore
//encouraged to load and test the software's suitability as regards their
//requirements in conditions enabling the security of their systems and/or
//data to be ensured and,  more generally, to use and operate it in the
//same conditions as regards security.
//
//The fact that you are presently reading this means that you have had
//knowledge of the CeCILL-B license and that you accept its terms.

/*
 *==============================================================================
 *==== example of multiple computations through getMSDQuaternion() function ====
 *==============================================================================
 */

#include "RapidRMSD.hpp"  // main library
#include "cPDBAtom.hpp" // for the PDB input
#include <stdlib.h>
#include <vector>

typedef double vector3[3]; 

// for the timer to evaluate elapsed time.
#if defined(_WIN32)
	#include <time.h>
	#include <Windows.h>

	#if defined(_MSC_VER) || defined(_MSC_EXTENSIONS)
		#define DELTA_EPOCH_IN_MICROSECS  11644473600000000Ui64
	#else
		#define DELTA_EPOCH_IN_MICROSECS  11644473600000000ULL
	#endif
 
	struct timezone 
	{
		int  tz_minuteswest; /* minutes W of Greenwich */
		int  tz_dsttime;     /* type of dst correction */
	};
 
	int gettimeofday(struct timeval *tv, struct timezone *tz)
	{
		FILETIME ft;
		unsigned __int64 tmpres = 0;
		static int tzflag;

		if (NULL != tv)
		{
			GetSystemTimeAsFileTime(&ft);
 
			tmpres |= ft.dwHighDateTime;
			tmpres <<= 32;
			tmpres |= ft.dwLowDateTime;
 
			/*converting file time to unix epoch*/
			tmpres -= DELTA_EPOCH_IN_MICROSECS; 
			tmpres /= 10;  /*convert into microseconds*/
			tv->tv_sec = (long)(tmpres / 1000000UL);
			tv->tv_usec = (long)(tmpres % 1000000UL);
		}
 
		if (NULL != tz)
		{
			if (!tzflag)
			{
				_tzset();
				tzflag++;
			}
			tz->tz_minuteswest = _timezone / 60;
			tz->tz_dsttime = _daylight;
		}
		return 0;
	}

#elif defined(__unix__) || ( defined(__APPLE__) && defined (__MACH__) )
#include <sys/time.h>	// gettimeofday is already there
#include <cmath>		// for sqrt
#endif


typedef double quat[4];

void mulQuat(quat &r, quat &a, quat &b) {

		r[3] = a[3]*b[3] - a[0]*b[0]  -  a[1]*b[1]  -  a[2]*b[2];
		r[0] = a[3]*b[0] + a[0]*b[3]  +  a[1]*b[2]  -  a[2]*b[1];
		r[1] = a[3]*b[1] + a[1]*b[3]  +  a[2]*b[0]  -  a[0]*b[2];
		r[2] = a[3]*b[2] + a[2]*b[3]  +  a[0]*b[1]  -  a[1]*b[0];

		return;
}

double getLinearRMSD(vector3 *coords, double* q, vector3 T, int N, double *weights=NULL) {

	double Mass = 0 ;
	double A = q[3]*q[3]-(q[0]*q[0]+q[1]*q[1]+q[2]*q[2]);
	
	double r = 0;
	double x,y,z;
	vector3 rtmp;
	if (weights) {
		for (int i=0; i<N; ++i) {

			Mass+=weights[i];

			x=coords[i][0];
			y=coords[i][1];
			z=coords[i][2];

			double B = (x*q[0]+y*q[1]+z*q[2]);
			double cx = q[1]*z-q[2]*y;
			double cy = q[2]*x-q[0]*z;
			double cz = q[0]*y-q[1]*x;

			rtmp[0] = A*x+2*q[3]*cx+2*B*q[0];
			rtmp[1] = A*y+2*q[3]*cy+2*B*q[1];
			rtmp[2] = A*z+2*q[3]*cz+2*B*q[2];

			rtmp[0]+=T[0];
			rtmp[1]+=T[1];
			rtmp[2]+=T[2];

			r+= weights[i]* ((x-rtmp[0])*(x-rtmp[0])+(y-rtmp[1])*(y-rtmp[1])+(z-rtmp[2])*(z-rtmp[2]));			
		}
		r/=Mass;
	}
	else {
		for (int i=0; i<N; ++i) {

			x=coords[i][0];
			y=coords[i][1];
			z=coords[i][2];

			double B = (x*q[0]+y*q[1]+z*q[2]);
			double cx = q[1]*z-q[2]*y;
			double cy = q[2]*x-q[0]*z;
			double cz = q[0]*y-q[1]*x;

			rtmp[0] = A*x+2*q[3]*cx+2*B*q[0];
			rtmp[1] = A*y+2*q[3]*cy+2*B*q[1];
			rtmp[2] = A*z+2*q[3]*cz+2*B*q[2];

			rtmp[0]+=T[0];
			rtmp[1]+=T[1];
			rtmp[2]+=T[2];

			r+= ((x-rtmp[0])*(x-rtmp[0])+(y-rtmp[1])*(y-rtmp[1])+(z-rtmp[2])*(z-rtmp[2]));			
		}
		r/=N;
	}
	return r;
}

int main() {

	std::cout<<"example of multiple computations through getMSDQuaternion() function\n";

	srand ( time(NULL) );

	// generate Nquat random orthonormal quaternions
	int Nquat = 25;
	quat* qArr = new quat[Nquat];
	double qLength;
	for (int i=0; i<Nquat; ++i) {
		qArr[i][0]= (rand()%100);
		qArr[i][1]= (rand()%100);
		qArr[i][2]= (rand()%100);
		qArr[i][3]= (rand()%100);
		qArr[i][0]/=100; qArr[i][1]/=100; qArr[i][2]/=100; qArr[i][3]/=100;
		//length of net quaternion
		qLength = sqrt(qArr[i][0]*qArr[i][0]+qArr[i][1]*qArr[i][1]+qArr[i][2]*qArr[i][2]+qArr[i][3]*qArr[i][3]);
		//normalize it
		qArr[i][0]/=qLength;
		qArr[i][1]/=qLength;
		qArr[i][2]/=qLength;
		qArr[i][3]/=qLength;
	}
	//translation:
	vector3 T;
	T[0] = (rand()%5);
	T[1] = (rand()%5);
	T[2] = (rand()%5);

	//test coordinates
	//initialization with pdbLines
	std::vector<cPDBAtom> atoms;
	atoms.resize(20);
	
	atoms[0].initialize("ATOM      4  N   GLU A   1       1.941  -2.006   3.760  1.00 25.22           N");
	atoms[1].initialize("ATOM      5  CA  GLU A   1       2.783  -2.083   2.553  1.00 29.88           C");
	atoms[2].initialize("ATOM      6  C   GLU A   1       2.110  -2.700   2.235  1.00 26.79           C");
	atoms[3].initialize("ATOM      7  O   GLU A   1       2.801  -2.237   1.337  1.00 22.16           O");
	atoms[4].initialize("ATOM      8  CB  GLU A   1       3.276  -2.495   2.288  1.00 25.90           C");
	atoms[5].initialize("ATOM      9  CG  GLU A   1       3.603  -0.188   3.574  1.00 48.92           C");
	atoms[6].initialize("ATOM     10  CD  GLU A   1       2.380  -0.852   4.025  1.00 29.36           C");
	atoms[7].initialize("ATOM     11  OE1 GLU A   1       1.978   1.935   3.576  1.00 28.67           O");
	atoms[8].initialize("ATOM     12  OE2 GLU A   1       1.821  -0.061   4.894  1.00 37.04           O");
	atoms[9].initialize("ATOM     13  N   TRP A   2       0.800  -2.015   2.074  1.00 20.72           N");
	atoms[10].initialize("ATOM     14  CA  TRP A   2       0.045  -3.649  -1.874  1.00 16.62           C");
	atoms[11].initialize("ATOM     15  C   TRP A   2      -0.264  -4.184  -1.771  1.00 18.21           C");
	atoms[12].initialize("ATOM     16  O   TRP A   2      -0.183  -5.585  -2.710  1.00 31.39           O");
	atoms[13].initialize("ATOM     17  CB  TRP A   2      -1.170  -2.520  -1.699  1.00 21.45           C");
	atoms[14].initialize("ATOM     18  CG  TRP A   2      -0.762  -1.955  -1.503  1.00 23.04           C");
	atoms[15].initialize("ATOM     19  CD1 TRP A   2      -0.834  -0.953   0.417  1.00 30.95           C");
	atoms[16].initialize("ATOM     20  CD2 TRP A   2      -0.289  -0.563  -2.315  1.00 18.69           C");
	atoms[17].initialize("ATOM     21  NE1 TRP A   2      -0.413  -2.140  -2.903  1.00 24.73           N");
	atoms[18].initialize("ATOM     22  CE2 TRP A   2      -0.203  -1.945  -3.564  1.00 27.52           C");
	atoms[19].initialize("ATOM     23  CE3 TRP A   2      -0.126  -0.093  -3.056  1.00 20.75           C");
	
	// compute coordinates and weights
	vector3 *coords = new vector3[atoms.size()];
	double *weights = new double[atoms.size()];
	
	for (unsigned int i=0; i<atoms.size(); i++) {
		coords[i][0] = atoms[i].coords[0];
		coords[i][1] = atoms[i].coords[1];
		coords[i][2] = atoms[i].coords[2];
		weights[i] = atoms[i].weight;
	}
	
	//initializing...
	RMSD<vector3> rmsd;
	rmsd.init(coords, atoms.size(), weights); 	// RMSD init

	// make Nquat^5 evaluations of rmsd...
	quat qtmp1, qtmp2, qtmp3, qNet;

	struct timeval _start, _end;
	double mtime, seconds, useconds; 
	gettimeofday(&_start, NULL);

	double MSD = 0;
	for (int i=0; i<Nquat; ++i) {
		for (int j=0; j<Nquat; ++j) {
			mulQuat(qtmp1, qArr[i], qArr[j]);
			for (int k=0; k<Nquat; ++k) {
				mulQuat (qtmp2,qtmp1, qArr[k]);
				for (int l=0; l<Nquat; ++l) {
					mulQuat (qtmp3, qtmp2, qArr[l]);
					for (int m=0; m<Nquat; ++m) {
						mulQuat (qNet, qtmp3, qArr[m]);
						MSD += rmsd.getMSDQuaternion(qNet, T); // RMSD function call
					}
				}
			}
		}
	}
	
	gettimeofday(&_end, NULL);

	seconds  = _end.tv_sec  - _start.tv_sec;
    useconds = _end.tv_usec - _start.tv_usec;

    mtime = ((seconds) * 1000.0 + useconds/1000.0) + 0.5;
	int nComputations = Nquat*Nquat*Nquat*Nquat*Nquat;
    printf("Elapsed time: %12lf seconds for %d MSD computations\n", double (mtime)*0.001, nComputations);
    printf("Or %12lf nanoseconds for 1 MSD computation\n", double (mtime)*1000*1000/nComputations);
	printf("Mean MSD = %lf\n", MSD/nComputations);
	
	delete [] coords;
	delete [] weights;
	delete [] qArr;

}
